from functools import lru_cache
from typing import Dict, Tuple, Union, List
import os
from transformers import AutoTokenizer, AutoConfig, AutoModel, BertModel, BertConfig
from transformers.optimization import get_linear_schedule_with_warmup, get_cosine_schedule_with_warmup
from lightning.pytorch import LightningModule
import torch
import tempfile
import json
from omegaconf import OmegaConf, DictConfig
from pathlib import Path
from torch.utils.data import DataLoader, BatchSampler, RandomSampler
from datasets import load_from_disk, load_dataset
from torch.optim.lr_scheduler import CyclicLR


def get_hf_tokenizer(config: Union[Dict, DictConfig] , vocab: Union[Dict, DictConfig]):
    with tempfile.TemporaryDirectory() as tmpdirname:
        vocab_path = os.path.join(tmpdirname, 'vocab.txt')
        config_path = os.path.join(tmpdirname, 'config.json')
        with open(vocab_path, 'w') as f:
            for k in vocab.keys():
                f.writelines(k + '\n')
        with open(config_path, 'w') as f:
            if type(config) == DictConfig:
                config = OmegaConf.to_container(config)
            assert type(config)==dict, f'config must be type dict, but found {type(config)}'
            d = json.dumps(dict(config))
            f.write(d)
        tokenizer = AutoTokenizer.from_pretrained(tmpdirname)
    return tokenizer


def get_hf_config_object(config: Union[DictConfig , Dict]) -> AutoConfig:
    with tempfile.TemporaryDirectory() as tmpdirname:
        config_path = os.path.join(tmpdirname, 'config.json')
        with open(config_path, 'w') as f:
            if type(config) == DictConfig:
                config = OmegaConf.to_container(config)
            try:
                config = config.to_dict() # 兼容huffingface config 对象
            except:
                pass
            assert type(config)==dict, f'config must be type dict, but found {type(config)}'
            d = json.dumps(dict(config))
            f.write(d)
        config = AutoConfig.from_pretrained(tmpdirname)
    return config

    
def align_token_span(token_span_offset: Tuple, token_offset_mapping: List[Tuple]) -> Tuple:
    '''将词符级别的下标对齐为字符级别的下标
    参数
    - token_span_offset: 例如(0, 1) 下标指的是字符的下标
    - token_offset_mapping: 每个词符与字符对应的下标[(0,1),(1,2)]
    返回
    char_span_offset: (0,2)
    '''
    char_span_offset = ()
    if token_span_offset[1] - token_span_offset[0] == 1:
        char_span_offset = token_offset_mapping[token_span_offset[0]]
        return char_span_offset

    else:
        start = token_offset_mapping[token_span_offset[0]][0]
        end = token_offset_mapping[token_span_offset[1]-1][1]
        char_span_offset = (start, end)
        return char_span_offset
    
class BaseModel(LightningModule):
    """最基础的模型类,配置了
    """
    scheduler_names = ['linear_warmup', 'cosine_warmup', 'harmonic', 'cycle']
    
    def get_linear_warmup_step_scheduler_config(self, optimizer) -> Dict:
        total_steps = self.get_total_steps()
        warmup_steps = self.get_one_epoch_steps() // 3
        scheduler = get_linear_schedule_with_warmup(optimizer=optimizer, num_training_steps=total_steps, num_warmup_steps=warmup_steps)
        scheduler_config = {'scheduler': scheduler, 'interval':'step'}
        return scheduler_config
    
    def get_cosine_warmup_step_scheduler_config(self, optimizer) -> Dict:
        total_steps = self.get_total_steps()
        warmup_steps = self.get_one_epoch_steps() // 3
        scheduler = get_cosine_schedule_with_warmup(optimizer=optimizer, num_training_steps=total_steps, num_warmup_steps=warmup_steps)
        scheduler_config = {'scheduler': scheduler, 'interval':'step'}
        return scheduler_config
    
    def get_harmonic_epoch_scheduler_config(self, optimizer):
        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        scheduler_config = {'scheduler': scheduler, 'interval':'epoch'}
        return scheduler_config
    
    def get_cycle_step_scheduler_config(self, optimizer):
        step_size_up = self.get_one_epoch_steps() // 2
        max_lr = self.hparams.lr
        base_lr = max_lr / 4
        scheduler = CyclicLR(optimizer=optimizer, base_lr=base_lr, max_lr=max_lr, step_size_up=step_size_up, cycle_momentum=False)
        scheduler_config = {'scheduler': scheduler, 'interval':'step'}
        return scheduler_config
        
    def get_total_steps(self):
        return self.trainer.estimated_stepping_batches
    
    def get_one_epoch_steps(self):
        total_steps = self.get_total_steps()
        max_epochs = self.trainer.max_epochs
        return total_steps // max_epochs
    
    def get_scheduler_config(self, optimizer, name: str):
        """

        Args:
            optimizer (): 优化器实例
            name (str): 'harmonic_epoch', 'linear_warmup_step', 'cosine_warmup_step'之一

        Returns:
            dict: scheduler配置字典
        """
        if name == 'harmonic':
            return self.get_harmonic_epoch_scheduler_config(optimizer=optimizer)
        elif name == 'linear_warmup':
            return self.get_linear_warmup_step_scheduler_config(optimizer=optimizer)
        elif name == 'cosine_warmup':
            return self.get_cosine_warmup_step_scheduler_config(optimizer=optimizer)
        elif name == 'cycle':
            return self.get_cycle_step_scheduler_config(optimizer=optimizer)
    
    
    
class PLMBaseModel(LightningModule):
    """基于预训练语言模型的基类,在继承类的时候需要传入plm和plm_dir
    
    - 内置了scheduler,可以通过cls.scheduler_names查看所有的scheduler,通过self.get_scheduler_config方法得到pl的scheduler config
    - 通过self.tokenizer直接调用tokenizer
    - 通过self.get_plm_architecture可以得到没有加载参数的预训练模型的架构
    """
    
    scheduler_names = ['linear_warmup', 'cosine_warmup', 'harmonic', 'cycle']
    
    def __init__(self) -> None:
        super().__init__()
        # 保存所有参数
        self.save_hyperparameters()
        assert self.hparams.scheduler in self.scheduler_names, f'availabel names {self.scheduler_names}'
        assert 'plm' in self.hparams and 'plm_dir' in self.hparams, 'you have to at least pass in plm and plm_dir'
          
    @property
    @lru_cache()
    def tokenizer(self):
        keys = self.hparams.keys()
        if 'trf_config' in keys and 'vocab' in keys:
            try:
                tokenizer = get_hf_tokenizer(self.hparams.trf_config, self.hparams.vocab)
            except:
                tokenizer = AutoTokenizer.from_pretrained(os.path.join(self.hparams.plm_dir, self.hparams.plm))
            return tokenizer
        elif 'plm' in keys and 'plm_dir' in keys:
            plm_path = Path(self.hparams.plm_dir, self.hparams.plm)
            if plm_path.exists():
                tokenizer = AutoTokenizer.from_pretrained(plm_path)
            else:
                tokenizer = AutoTokenizer.from_pretrained(self.hparams.plm)
            return tokenizer

    @property
    @lru_cache()
    def trf_config(self) -> AutoConfig:
        if 'trf_config' in self.hparams.keys():
            return get_hf_config_object(self.hparams.trf_config)
        elif 'plm' in self.hparams.keys() and 'plm_dir' in self.hparams.keys():
            plm_path = os.path.join(self.hparams.plm_dir, self.hparams.plm)
            plm_config = AutoConfig.from_pretrained(plm_path) 
            self.hparams.trf_config = plm_config.to_dict() 
            self.hparams.vocab = dict(sorted(self.tokenizer.vocab.items(), key=lambda x: x[1]))  
            return plm_config
    
    def get_plm_architecture(self, add_pooler_layer: bool = False) -> torch.nn.Module:
        trf_config = self.trf_config
        trf_config.add_pooler_layer = add_pooler_layer
        return AutoModel.from_config(trf_config)    
    
    def get_linear_warmup_step_scheduler_config(self, optimizer) -> Dict:
        total_steps = self.get_total_steps()
        warmup_steps = self.get_one_epoch_steps() // 3
        scheduler = get_linear_schedule_with_warmup(optimizer=optimizer, num_training_steps=total_steps, num_warmup_steps=warmup_steps)
        scheduler_config = {'scheduler': scheduler, 'interval':'step'}
        return scheduler_config
    
    def get_cosine_warmup_step_scheduler_config(self, optimizer) -> Dict:
        total_steps = self.get_total_steps()
        warmup_steps = self.get_one_epoch_steps() // 3
        scheduler = get_cosine_schedule_with_warmup(optimizer=optimizer, num_training_steps=total_steps, num_warmup_steps=warmup_steps)
        scheduler_config = {'scheduler': scheduler, 'interval':'step'}
        return scheduler_config
    
    def get_harmonic_epoch_scheduler_config(self, optimizer):
        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        scheduler_config = {'scheduler': scheduler, 'interval':'epoch'}
        return scheduler_config
    
    def get_cycle_step_scheduler_config(self, optimizer):
        step_size_up = self.get_one_epoch_steps() // 2
        max_lr = self.hparams.lr
        base_lr = max_lr / 4
        scheduler = CyclicLR(optimizer=optimizer, base_lr=base_lr, max_lr=max_lr, step_size_up=step_size_up, cycle_momentum=False)
        scheduler_config = {'scheduler': scheduler, 'interval':'step'}
        return scheduler_config
        
    def get_total_steps(self):
        return self.trainer.estimated_stepping_batches
    
    def get_one_epoch_steps(self):
        total_steps = self.get_total_steps()
        max_epochs = self.trainer.max_epochs
        return total_steps // max_epochs
        
    def get_scheduler_config(self, optimizer, name: str):
        """

        Args:
            optimizer (): 优化器实例
            name (str): 'harmonic_epoch', 'linear_warmup_step', 'cosine_warmup_step'之一

        Returns:
            dict: scheduler配置字典
        """
        if name == 'harmonic':
            return self.get_harmonic_epoch_scheduler_config(optimizer=optimizer)
        elif name == 'linear_warmup':
            return self.get_linear_warmup_step_scheduler_config(optimizer=optimizer)
        elif name == 'cosine_warmup':
            return self.get_cosine_warmup_step_scheduler_config(optimizer=optimizer)
        elif name == 'cycle':
            return self.get_cycle_step_scheduler_config(optimizer=optimizer)

    def to_onnx(self, 
                file_path: str, 
                text: str = '中国人'):
        torch_inputs = self.tokenizer(text, return_tensors='pt')
        dynamic_axes = {
                    'input_ids': {0: 'batch', 1: 'seq'},
                    'attention_mask': {0: 'batch', 1: 'seq'},
                    'token_type_ids': {0: 'batch', 1: 'seq'},
                }
        with torch.no_grad():
            torch.onnx.export(model=self,
                              args=tuple(torch_inputs.values()), 
                              f=file_path, 
                              input_names=list(torch_inputs.keys()),
                              dynamic_axes=dynamic_axes, 
                              opset_version=14,
                              output_names=['logits'],
                              export_params=True)
        print('export to onnx successfully')
        

class HFPretrainedModel(BaseModel):
    """配置了所有功能的模型基类
    """
    
    def __init__(self,
                 plm_max_length: int = 512,
                 num_workers: int = 4,
                 pin_memory: bool = True,
                 drop_last_batch: bool = False,
                 shuffle_train: bool = False,
                 shuffle_val: bool = False,
                 shuffle_test: bool = False):
        super().__init__()
        # 保存所有参数
        self.save_hyperparameters()
        assert 'plm' in self.hparams, 'you have to at least pass in plm'
        
    @property
    @lru_cache()
    def dataset(self):
        path = Path(self.hparams.dataset)
        if path.exists():
            return load_from_disk(path)
        else:
            return load_dataset(path=self.hparams.dataset)
          
    @property
    @lru_cache()
    def tokenizer(self):
        keys = self.hparams.keys()
        if 'trf_config' in keys and 'vocab' in keys:
            return get_hf_tokenizer(self.hparams.trf_config, self.hparams.vocab)
        elif 'plm' in keys:
            plm_path = Path(self.hparams.plm)
            tokenizer = AutoTokenizer.from_pretrained(plm_path)
            return tokenizer

    def get_plm_config(self):
        if 'trf_config' in self.hparams.keys():
            return get_hf_config_object(self.hparams.trf_config)
        elif 'plm' in self.hparams.keys():
            plm_config = AutoConfig.from_pretrained(self.hparams.plm)    
            return plm_config
        
    def get_hf_config_object(self, config: Union[DictConfig , Dict]) -> AutoConfig:
        with tempfile.TemporaryDirectory() as tmpdirname:
            config_path = os.path.join(tmpdirname, 'config.json')
            with open(config_path, 'w') as f:
                if type(config) == DictConfig:
                    config = OmegaConf.to_container(config)
                try:
                    config = config.to_dict() # 兼容huffingface config 对象
                except:
                    pass
                assert type(config)==dict, f'config must be type dict, but found {type(config)}'
                d = json.dumps(dict(config))
                f.write(d)
            config = AutoConfig.from_pretrained(tmpdirname)
        return config
    
    def get_plm(self) -> torch.nn.Module:
        """获取预训练模型,用于初始化模型
        """
        if 'trf_config' in self.hparams.keys():
            plm_config = self.get_hf_config_object(self.hparams.trf_config)
            plm = AutoModel(plm_config)
        elif 'plm' in self.hparams.keys():
            plm = AutoModel.from_pretrained(self.hparams.plm)
            self.hparams.trf_config = plm.config.to_dict()
            self.hparams.vocab = dict(sorted(self.tokenizer.vocab.items(), key=lambda x: x[1]))
        return plm   

    def to_onnx(self, 
                file_path: str, 
                text: str = '中国人'):
        torch_inputs = self.tokenizer(text, return_tensors='pt')
        dynamic_axes = {
                    'input_ids': {0: 'batch', 1: 'seq'},
                    'attention_mask': {0: 'batch', 1: 'seq'},
                    'token_type_ids': {0: 'batch', 1: 'seq'},
                }
        with torch.no_grad():
            torch.onnx.export(model=self,
                              args=tuple(torch_inputs.values()), 
                              f=file_path, 
                              input_names=list(torch_inputs.keys()),
                              dynamic_axes=dynamic_axes, 
                              opset_version=14,
                              output_names=['logits'],
                              export_params=True)
        print('export to onnx successfully')
        
    
    @property
    @lru_cache()
    def train_df(self):
        return self.dataset['train'].to_pandas()
    
    @property
    @lru_cache()
    def val_df(self):
        return self.dataset['validation'].to_pandas()
    
    @property
    @lru_cache()
    def test_df(self):
        return self.dataset['test'].to_pandas()
    
    def train_dataloader(self):
        return DataLoader(dataset= self.dataset['train'], 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=self.hparams.shuffle_train,
                          batch_size=None,
                          sampler=BatchSampler(RandomSampler(self.dataset['train']), batch_size=self.hparams.batch_size, drop_last=self.hparams.drop_last_batch))
    
    def val_dataloader(self):
        return DataLoader(dataset=self.dataset['validation'], 
                          batch_size=None, 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=self.hparams.shuffle_val,
                          sampler=BatchSampler(RandomSampler(self.dataset['validation']), batch_size=self.hparams.batch_size, drop_last=self.hparams.drop_last_batch))

    def test_dataloader(self):
        return DataLoader(dataset=self.dataset['test'], 
                          batch_size=None, 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=self.hparams.shuffle_test,
                          sampler=BatchSampler(RandomSampler(self.dataset['test']), batch_size=self.hparams.batch_size, drop_last=self.hparams.drop_last_batch))



class HFBertModel(BaseModel):
    """配置了所有功能的模型基类
    """
    
    def __init__(self,
                 plm_dir: str,
                 input_max_length: int = 512):
        super().__init__()
        # 保存所有参数
        self.save_hyperparameters()
        self.input_max_length = input_max_length
        self.bert = self.get_plm()
        
    def on_train_start(self) -> None:
        ckpt_path = Path(self.hparams.plm_dir, 'pytorch_model.bin')
        if ckpt_path.exists():
            self.bert.load_state_dict(torch.load(ckpt_path))
        else:
            self.bert = self.bert.from_pretrained(self.hparams.plm_dir)
        
    @property
    @lru_cache()
    def tokenizer(self):
        keys = self.hparams.keys()
        if 'trf_config' in keys and 'vocab' in keys:
            return get_hf_tokenizer(self.hparams.trf_config, self.hparams.vocab)
        elif 'plm' in keys:
            plm_path = Path(self.hparams.plm_dir)
            tokenizer = AutoTokenizer.from_pretrained(plm_path)
            return tokenizer

    def get_plm_config(self):
        if 'trf_config' in self.hparams.keys():
            return get_hf_config_object(self.hparams.trf_config)
        elif 'plm' in self.hparams.keys():
            plm_config = BertConfig.from_pretrained(self.hparams.plm_dir)    
            return plm_config
        
    def get_hf_config_object(self, config: Union[DictConfig , Dict]) -> AutoConfig:
        with tempfile.TemporaryDirectory() as tmpdirname:
            config_path = os.path.join(tmpdirname, 'config.json')
            with open(config_path, 'w') as f:
                if type(config) == DictConfig:
                    config = OmegaConf.to_container(config)
                try:
                    config = config.to_dict() # 兼容huffingface config 对象
                except:
                    pass
                assert type(config)==dict, f'config must be type dict, but found {type(config)}'
                d = json.dumps(dict(config))
                f.write(d)
            config = AutoConfig.from_pretrained(tmpdirname)
        return config
    
    def get_plm(self) -> torch.nn.Module:
        """获取预训练模型,用于初始化模型
        """
        if 'trf_config' in self.hparams.keys():
            plm_config = self.get_hf_config_object(self.hparams.trf_config)
            plm = BertModel(plm_config)
        elif 'plm' in self.hparams.keys():
            plm = BertModel.from_pretrained(self.hparams.plm_dir)
            self.hparams.trf_config = plm.config.to_dict()
            self.hparams.vocab = dict(sorted(self.tokenizer.vocab.items(), key=lambda x: x[1]))
        return plm   

    def to_onnx(self, 
                file_path: str, 
                text: str = '中国人'):
        torch_inputs = self.tokenizer(text, return_tensors='pt')
        dynamic_axes = {
                    'input_ids': {0: 'batch', 1: 'seq'},
                    'attention_mask': {0: 'batch', 1: 'seq'},
                    'token_type_ids': {0: 'batch', 1: 'seq'},
                }
        with torch.no_grad():
            torch.onnx.export(model=self,
                              args=tuple(torch_inputs.values()), 
                              f=file_path, 
                              input_names=list(torch_inputs.keys()),
                              dynamic_axes=dynamic_axes, 
                              opset_version=14,
                              output_names=['logits'],
                              export_params=True)
        print('export to onnx successfully')


import pytorch_lightning as pl 
from transformers import BertModel, AutoConfig, AutoTokenizer
import torch
import os
from torchmetrics import SpearmanCorrCoef
from ...layers.loss import CoSentLoss
from rich.progress import track



class CoSentBERT(pl.LightningModule):
    '''更好的句向量方案CoSent实现
    参考:
    - https://blog.csdn.net/HUSTHY/article/details/122821034?spm=1001.2101.3001.6650.1&utm_medium=distribute.pc_relevant.none-task-blog-2%7Edefault%7ECTRLIST%7Edefault-1-122821034-blog-124220249.pc_relevant_aa&depth_1-utm_source=distribute.pc_relevant.none-task-blog-2%7Edefault%7ECTRLIST%7Edefault-1-122821034-blog-124220249.pc_relevant_aa&utm_relevant_index=2
    '''
    def __init__(self, 
                 lr: float,
                 weight_decay: float = 0.0,
                 **kwargs):
        """

        Args:
            lr (float): 学习率
            weight_decay (float): 权重衰减,默认为0
        """
        super().__init__()
        self.save_hyperparameters(logger=False)

        self.bert = BertModel(self.hparams.trf_config)
        self.tokenizer = self._init_tokenizer()
        self.criterion = CoSentLoss()
        self.train_meric = SpearmanCorrCoef()
        self.val_metric = SpearmanCorrCoef()
        self.test_metric = SpearmanCorrCoef()
        
    def forward(self, 
                input_ids, 
                token_type_ids, 
                attention_mask):
        hidden_state = self.bert(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask).last_hidden_state
        hidden_state = torch.sum(hidden_state * attention_mask[:, :, None], dim=1)
        attention_mask = torch.sum(attention_mask, dim=1)[:, None]
        return hidden_state / attention_mask
    
    def step(self, batch):
        inputs_a = batch['inputs_a']
        inputs_b = batch['inputs_b']
        targs = batch['similarities']
        outputs_a = self(**inputs_a)
        outputs_b = self(**inputs_b)
        preds = torch.cosine_similarity(outputs_a, outputs_b)
        loss = self.criterion(preds, targs, self.device)
        return loss, preds, targs
    
    def training_step(self, batch, batch_idx):
        loss, preds, targs = self.step(batch)
        score = self.train_meric(preds, targs)
        self.log('train/spearman', self.train_meric, on_step=True, prog_bar=True)
        self.log('train/loss', loss)
        return {'loss': loss}

    def validation_step(self, batch, batch_idx):
        loss, preds, targs = self.step(batch)
        score = self.val_metric(preds, targs)
        self.log('val/spearman', self.val_metric, on_epoch=True, prog_bar=True)
        return {'loss': loss}
    
    def test_step(self, batch, batch_idx):
        loss, preds, targs = self.step(batch)
        self.test_metric(preds, targs)
        self.log('test/spearman', self.test_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}

    def configure_optimizers(self):
        self.optimizer = torch.optim.Adam(self.parameters(), lr=self.hparams.lr, weight_decay=self.hparams.weight_decay)
        self.scheduler = torch.optim.lr_scheduler.LambdaLR(self.optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        return [self.optimizer], [self.scheduler]

    def _init_tokenizer(self):
        with open('./vocab.txt', 'w') as f:
            for k in self.hparams.vocab.keys():
                f.writelines(k + '\n')
        self.hparams.trf_config.to_json_file('./config.json')
        tokenizer = AutoTokenizer.from_pretrained('./')
        os.remove('./vocab.txt')
        os.remove('./config.json')
        return tokenizer
    
    def encode(self, texts: list, device: str = 'cpu', batch_size: int = 500):
        device = torch.device(device)
        self.to(device)
        self.eval()
        all_embeds = torch.tensor([], device=device)
        with torch.no_grad():
            for i in track(range(0, len(texts), batch_size), description='Encoding'):
                batch = self.tokenizer(texts[i:i+batch_size], padding='max_length', return_tensors='pt', max_length=self.hparams.max_length, truncation=True)
                batch = {k: v.to(device) for k, v in batch.items()}
                embeds = self(**batch)
                all_embeds = torch.cat((all_embeds, embeds), dim=0)
        return all_embeds
    
    def predict(self, text_a: str, text_b: str, device: str = 'cpu'):
        device = torch.device(device)
        self.to(device)
        self.eval()
        with torch.no_grad():
            embed_a = self.tokenizer(text_a, padding='max_length', truncation=True, max_length=self.max_length)
            embed_b = self.tokenizer(text_b, padding='max_length', truncation=True, max_length=self.max_length)
            cos_sim = torch.cosine_similarity(embed_a, embed_b)
        return cos_sim


from .attention import MultiHeadAttentionLayer
from .normalization import LayerNorm as BertLayerNorm
import torch.nn as nn
import torch
from .normalization import LayerNorm
import os
from typing import Optional,  Tuple
from .activation import activations


class BertEmbeddings(nn.Module):
    """bert的Embedding"""
    def __init__(
        self,
        vocab_size,
        type_vocab_size,
        hidden_size,
        max_position_embeddings,
        hidden_dropout_prob,
        layer_norm_eps
        ) -> None:
        super().__init__()
        self.word_embeddings = nn.Embedding(vocab_size, hidden_size)
        self.position_embeddings = nn.Embedding(max_position_embeddings, hidden_size)
        self.token_type_embeddings = nn.Embedding(type_vocab_size, hidden_size)
        self.dropout = nn.Dropout(hidden_dropout_prob)
        self.layer_norm = LayerNorm(hidden_size=hidden_size, eps=layer_norm_eps)


    def forward(self, input_ids, token_type_ids=None, position_ids=None):
        seq_length = input_ids.size(1)
        if position_ids is None:
            position_ids = torch.arange(seq_length, dtype=torch.long, device=input_ids.device)
            position_ids = position_ids.unsqueeze(0).expand_as(input_ids)
        # 如果不传token_type 则默认为不需要做序列种类判断
        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)
        word_embeddings = self.word_embeddings(input_ids)
        position_embeddings = self.position_embeddings(position_ids)
        token_type_embeddings = self.token_type_embeddings(token_type_ids)
        embeddings = word_embeddings + position_embeddings + token_type_embeddings
        embeddings = self.layer_norm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings


class BertIntermediate(nn.Module):
    def __init__(self, hidden_size, intermediate_size, hidden_act) -> None:
        super().__init__()
        self.dense = nn.Linear(hidden_size, intermediate_size)
        self.intermediate_act_fn = activations[hidden_act]

    def forward(self, hidden_states):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.intermediate_act_fn(hidden_states)
        return hidden_states


class BertAddNorm(nn.Module):
    def __init__(self, intermediate_size, hidden_size, hidden_dropout_prob, layer_norm_eps) -> None:
        super().__init__()
        self.dense = nn.Linear(intermediate_size, hidden_size)
        self.dropout = nn.Dropout(hidden_dropout_prob)
        self.layer_norm = BertLayerNorm(hidden_size=hidden_size, eps=layer_norm_eps)

    def forward(self, hidden_states, input_tensor):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.layer_norm(hidden_states + input_tensor)
        return hidden_states

class BertAttention(nn.Module):
    def __init__(
        self, 
        hidden_size, 
        num_attention_heads, 
        attention_probs_dropout_prob, 
        hidden_dropout_prob,
        layer_norm_eps) -> None:
        super().__init__()
        self.self = MultiHeadAttentionLayer(hidden_size, num_attention_heads, attention_probs_dropout_prob)
        self.output = BertAddNorm(
            intermediate_size=hidden_size,
            hidden_size=hidden_size,
            hidden_dropout_prob=hidden_dropout_prob,
            layer_norm_eps=layer_norm_eps
            )
    def forward(self, input_tensor, attention_mask=None, head_mask=None):
        self_outputs = self.self(input_tensor, input_tensor, input_tensor, attention_mask, head_mask)
        attention_output = self.output(self_outputs[0], input_tensor)
        outputs = (attention_output,) + self_outputs[1:]
        return outputs


class BertLayer(nn.Module):
    def __init__(
        self,
        hidden_size,
        num_attention_heads,
        hidden_act,
        intermediate_size,
        hidden_dropout_prob,
        layer_norm_eps
        ) -> None:
        super().__init__()
        self.attention = BertAttention(
            hidden_size=hidden_size,
            num_attention_heads=num_attention_heads,
            attention_probs_dropout_prob=hidden_dropout_prob,
            hidden_dropout_prob=hidden_dropout_prob,
            layer_norm_eps=layer_norm_eps
        )
        self.intermediate = BertIntermediate(hidden_size=hidden_size,intermediate_size=intermediate_size, hidden_act=hidden_act)
        self.output = BertAddNorm(intermediate_size, hidden_size, hidden_dropout_prob, layer_norm_eps)

    def forward(self, hidden_states, attention_mask=None, head_mask=None):
        attention_outputs = self.attention(hidden_states, attention_mask, head_mask)
        attention_output = attention_outputs[0]

        # 这里是左上的 Add & Norm，从而得到完整的 FFN
        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output)

        # attention_outputs[0] 是 embedding, [1] 是 attention_probs
        outputs = (layer_output,) + attention_outputs[1:]
        return outputs


class BertEncoder(nn.Module):
    def __init__(self,
                 num_hidden_layers,
                 hidden_size,
                 num_attention_heads,
                 hidden_act,
                 intermediate_size,
                 hidden_dropout_prob,
                 layer_norm_eps,
                 output_attentions,
                 output_hidden_states) -> None:
        super().__init__()
        self.layer = nn.ModuleList([BertLayer(hidden_size, num_attention_heads, hidden_act, intermediate_size, hidden_dropout_prob, layer_norm_eps) for _ in range(num_hidden_layers)])
        self.output_attentions = output_attentions
        self.output_hidden_states = output_hidden_states


    def forward(self, hidden_states, attention_mask=None, head_mask=None):
        all_hidden_states = ()
        all_attentions = ()
        for i, layer_module in enumerate(self.layer):
            if self.output_hidden_states:
                all_hidden_states += (hidden_states, )
            layer_outputs = layer_module(hidden_states, attention_mask, head_mask[i])
            hidden_states = layer_outputs[0]
            if self.output_attentions:
                all_attentions += (layer_outputs[1], )
        
        # Add last layer
        if self.output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        outputs = (hidden_states,)
        if self.output_hidden_states:
            # 把中间层的结果取出来，一些研究认为中间层的 embedding 也有价值
            outputs = outputs + (all_hidden_states,)
        if self.output_attentions:
            outputs = outputs + (all_attentions,)

        # last-layer hidden state, (all hidden states), (all attentions)
        return outputs


class BertPooler(nn.Module):
    def __init__(
        self,
        hidden_size) -> None:
        super().__init__()
        self.dense = nn.Linear(hidden_size, hidden_size)
        self.activation = nn.Tanh()

    def forward(self, hidden_states):
        cls = hidden_states[:, 0]
        pooled_output = self.dense(cls)
        pooled_output = self.activation(pooled_output)
        return pooled_output


class BertOutput:
    last_hidden_state: torch.FloatTensor = None
    pooler_output: torch.FloatTensor = None
    attentions: Optional[Tuple[torch.FloatTensor]] = None

    def __init__(self, last_hidden_state, pooler_output, attentions):
        self.last_hidden_state = last_hidden_state
        self.pooler_output = pooler_output
        self.attentions = attentions



class Bert(nn.Module):
    def __init__(self,
        vocab_size: int,
        type_vocab_size: int =2,
        max_position_embeddings: int = 512,
        hidden_size: int = 768,
        hidden_dropout_prob: float = 0.1,
        num_attention_heads: int = 8,
        num_hidden_layers: int = 12,
        hidden_act: str = 'gelu',
        intermediate_size: int = 3072,
        output_attentions: bool = False,
        output_hidden_states :bool = False,
        layer_norm_eps:float = 3e-12
        ) -> None:
        super().__init__()
        self.embeddings =BertEmbeddings(
            vocab_size=vocab_size,
            type_vocab_size=type_vocab_size,
            hidden_size=hidden_size,
            max_position_embeddings=max_position_embeddings,
            hidden_dropout_prob=hidden_dropout_prob,
            layer_norm_eps=layer_norm_eps
        )
        self.encoder = BertEncoder(
            hidden_size=hidden_size,
            num_hidden_layers=num_hidden_layers,
            num_attention_heads=num_attention_heads,
            hidden_act=hidden_act,
            intermediate_size=intermediate_size,
            hidden_dropout_prob=hidden_dropout_prob,
            layer_norm_eps=layer_norm_eps,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states
        )
        self.pooler = BertPooler(hidden_size=hidden_size)


    def init_weights(self):
        self.apply(self._init_weights)

    def _init_weights(self, module):
        """ Initialize the weights """
        if isinstance(module, (nn.Linear, nn.Embedding)):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
        elif isinstance(module, BertLayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        if isinstance(module, nn.Linear) and module.bias is not None:
            module.bias.data.zero_()

    def from_pretrained(self, pretrained_model_path):
        if not os.path.exists(pretrained_model_path):
            print(f"missing pretrained_model_path: {pretrained_model_path}")
            pass

        state_dict = torch.load(pretrained_model_path, map_location='cpu')

        # 名称可能存在不一致，进行替换
        old_keys = []
        new_keys = []
        for key in state_dict.keys():
            new_key = key
            if 'gamma' in key:
                new_key = new_key.replace('gamma', 'weight')
            if 'beta' in key:
                new_key = new_key.replace('beta', 'bias')
            if 'bert.' in key:
                new_key = new_key.replace('bert.', '')
            # 兼容部分不优雅的变量命名
            if 'LayerNorm' in key:
                new_key = new_key.replace('LayerNorm', 'layer_norm')


            if new_key:
                old_keys.append(key)
                new_keys.append(new_key)

        for old_key, new_key in zip(old_keys, new_keys):

            if new_key in self.state_dict().keys():
                state_dict[new_key] = state_dict.pop(old_key)
            else:
                # 避免预训练模型里有多余的结构，影响 strict load_state_dict
                state_dict.pop(old_key)

        # 确保完全一致
        self.load_state_dict(state_dict, strict=True)

    def forward(self, input_ids, attention_mask=None, token_type_ids=None, position_ids=None, head_mask=None):
        if attention_mask is None:
            attention_mask = torch.ones_like(input_ids)
        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)

        # We create a 3D attention mask from a 2D tensor mask.
        # Sizes are [batch_size, 1, 1, to_seq_length]
        # So we can broadcast to [batch_size, num_heads, from_seq_length, to_seq_length]
        # this attention mask is more simple than the triangular masking of causal attention
        # used in OpenAI GPT, we just need to prepare the broadcast dimension here.

        extended_attention_mask = attention_mask.unsqueeze(1).unsqueeze(2)

        # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
        # masked positions, this operation will create a tensor which is 0.0 for
        # positions we want to attend and -10000.0 for masked positions.
        # Since we are adding it to the raw scores before the softmax, this is
        # effectively the same as removing these entirely.
        extended_attention_mask = extended_attention_mask.to(dtype=next(self.parameters()).dtype)  # fp16 compatibility
        extended_attention_mask = (1.0 - extended_attention_mask) * -10000.0

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x seq_length x seq_length]
        if head_mask is not None:
            if head_mask.dim() == 1:
                head_mask = head_mask.unsqueeze(0).unsqueeze(0).unsqueeze(-1).unsqueeze(-1)
                head_mask = head_mask.expand(self.config.num_hidden_layers, -1, -1, -1, -1)
            elif head_mask.dim() == 2:
                head_mask = head_mask.unsqueeze(1).unsqueeze(-1).unsqueeze(
                    -1)  # We can specify head_mask for each layer
            head_mask = head_mask.to(
                dtype=next(self.parameters()).dtype)  # switch to fload if need + fp16 compatibility
        else:
            head_mask = [None] * self.config.num_hidden_layers

        embedding_output = self.embeddings(input_ids, position_ids=position_ids, token_type_ids=token_type_ids)
        encoder_outputs = self.encoder(embedding_output,
                                       extended_attention_mask,
                                       head_mask=head_mask)
        sequence_output = encoder_outputs[0]
        pooled_output = self.pooler(sequence_output)

        outputs = BertOutput(last_hidden_state=sequence_output, pooler_output=pooled_output,
                             attentions=encoder_outputs[1:])

        return outputs
    
    
class BertBaseConfig:
    attention_probs_dropout_prob = 0.1,
    gradient_checkpointing= False,
    hidden_act = "gelu",
    hidden_dropout_prob = 0.1,
    hidden_size= 768,
    initializer_range= 0.02,
    intermediate_size= 3072,
    layer_norm_eps= 1e-12,
    max_position_embeddings= 512,
    num_attention_heads= 12,
    num_hidden_layers= 12,
    pad_token_id= 0,
    pooler_fc_size= 768,
    pooler_num_attention_heads= 12,
    pooler_num_fc_layers= 3,
    pooler_size_per_head= 128,
    position_embedding_type= 'absolute',
    type_vocab_size= 2,
    use_cache= True,
    vocab_size= 21128


from ...layers import MultiDropout, EfficientBiaffineSpanClassifier
from ...layers.loss import SparseMultiLabelCrossEntropy
from ...metrics.relation import RelationF1, Relation, Entity
from ...metrics.span import SpanF1
import torch
from torch import Tensor
from typing import List, Set
from ...utils.make_model import align_token_span, PLMBaseModel
from typing import Optional


class BLinkerForEntityRelationExtraction(PLMBaseModel):
    """基于biaffine的实体关系联合抽取模型
    """
    def __init__(self,
                 hidden_size: int = 64,
                 lr: float = 3e-5,
                 scheduler: str = 'linear_warmup', 
                 weight_decay: float = 0.01,
                 threshold: float = 0.0,
                 **kwargs):
        super().__init__()
        
        self.plm = self.get_plm_architecture()
        self.dropout = MultiDropout()       
        # 主语 宾语分类器
        self.ent_classifier = EfficientBiaffineSpanClassifier(self.plm.config.hidden_size, hidden_size, len(self.hparams.id2combined))
        # 主语 宾语 头对齐
        self.head_classifier = EfficientBiaffineSpanClassifier(self.plm.config.hidden_size, hidden_size, len(self.hparams.id2rel), add_rope=False, tril_mask=False)
        # 主语 宾语 尾对齐
        self.tail_classifier = EfficientBiaffineSpanClassifier(self.plm.config.hidden_size, hidden_size, len(self.hparams.id2rel), add_rope=False, tril_mask=False)

        self.ent_criterion = SparseMultiLabelCrossEntropy()
        self.head_criterion = SparseMultiLabelCrossEntropy()
        self.tail_criterion = SparseMultiLabelCrossEntropy()
        

        self.ent_metric = SpanF1()
        self.head_metric = SpanF1()
        self.tail_metric = SpanF1()
        self.val_metric = RelationF1()
        
        
    def setup(self, stage: str) -> None:
        self.trainer.datamodule.dataset['train'].set_transform(self.trainer.datamodule.sparse_combined_transform)
        self.trainer.datamodule.dataset['validation'].set_transform(self.trainer.datamodule.combined_transform)


    def forward(self, input_ids, attention_mask=None):
        hidden_state = self.plm(input_ids=input_ids, attention_mask=attention_mask).last_hidden_state
        hidden_state = self.dropout(hidden_state)
        ent_logits = self.ent_classifier(hidden_state, mask=attention_mask)
        head_logits = self.head_classifier(hidden_state, mask=attention_mask)
        tail_logits = self.tail_classifier(hidden_state, mask=attention_mask)
        return ent_logits, head_logits, tail_logits


    def shared_step(self, batch, is_train: bool):
        #inputs为bert常规输入, span_ids: [batch_size, 2, seq_len, seq_len],
        #head_ids: [batch_size, len(label2id), seq_len, seq_len], tail_ids: [batch_size, len(label2id), seq_len, seq_len]
        input_ids, attention_mask = batch['input_ids'], batch['attention_mask']
        ent_true, head_true, tail_true = batch['combined_tags'],  batch['head_tags'], batch['tail_tags']
        ent_logits, head_logits, tail_logits = self(input_ids=input_ids, attention_mask=attention_mask)
        
        if is_train:
            b,c,s,s = ent_logits.shape
            ent_logits = ent_logits.reshape(b,c,s*s)
            ent_true = ent_true[..., 0] * s + ent_true[..., 1]
            ent_loss = self.ent_criterion(ent_logits, ent_true)
        
            b,c,s,s = head_logits.shape
            head_logits = head_logits.reshape(b,c,s*s)
            head_true = head_true[..., 0] * s + head_true[..., 1]
            head_loss = self.head_criterion(head_logits, head_true)
            
            b,c,s,s = tail_logits.shape
            tail_logits = tail_logits.reshape(b,c,s*s)
            tail_true = tail_true[..., 0] * s + tail_true[..., 1]
            tail_loss = self.tail_criterion(tail_logits, tail_true)
            self.log('train/ent_loss', ent_loss, prog_bar=True, on_step=True)
            self.log('train/head_loss', head_loss, prog_bar=True, on_step=True)
            self.log('train/tail_loss', tail_loss, prog_bar=True, on_step=True)
            
            loss = sum([ent_loss, head_loss, tail_loss]) / 3
            return loss
        else:
            return ent_logits, head_logits, tail_logits, ent_true, head_true, tail_true

        
    def training_step(self, batch, batch_idx) -> dict:
        # 训练阶段不进行解码, 会比较慢
        loss = self.shared_step(batch, is_train=True)
        return {'loss': loss}


    def validation_step(self, batch, batch_idx) -> dict:
        ent_logits, head_logits, tail_logits, ent_true, head_true, tail_true = self.shared_step(batch, is_train=False)
        
        ent_pred = ent_logits.gt(self.hparams.threshold)
        head_pred = head_logits.gt(self.hparams.threshold)
        tail_pred = tail_logits.gt(self.hparams.threshold)
        
        self.ent_metric(ent_pred, ent_true)
        self.log('val/ent', self.ent_metric, on_epoch=True, prog_bar=True)

        self.head_metric(head_pred, head_true)
        self.log('val/head', self.head_metric, on_epoch=True, prog_bar=True)

        self.tail_metric(tail_pred, tail_true)
        self.log('val/tail', self.tail_metric, on_epoch=True, prog_bar=True)
        
        batch_rels = self.extract_rels(ent_logits, head_logits, tail_logits, threshold=self.hparams.threshold)
        batch_true_rels = self.extract_rels(ent_true, head_true, tail_true, threshold=self.hparams.threshold)
        self.val_metric(batch_rels, batch_true_rels)
        self.log('val/f1', self.val_metric, on_epoch=True, prog_bar=True)


    def configure_optimizers(self)  :
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.named_parameters() if not any(nd in n for nd in no_decay)],
            'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.named_parameters() if any(nd in n for nd in no_decay)],
            'weight_decay': 0.0},
        ]
        optimizer = torch.optim.AdamW(grouped_parameters, lr=self.hparams.lr)
        scheduler_config = self.get_scheduler_config(optimizer=optimizer, name=self.hparams.scheduler)
        return [optimizer], [scheduler_config]



    def extract_rels(self, 
                     ent_logits: Tensor,
                     head_logits: Tensor, 
                     tail_logtis: Tensor, 
                     threshold: Optional[float] = None,
                     mapping: Optional[dict] = None) -> List[Set[Relation]]:
        """
        将三个globalpointer预测的结果进行合并，得到三元组的预测结果
        参数:
        - ent_logits: [batch_size, 2, seq_len, seq_len]
        - head_logits: [batch_size, predicate_type, seq_len, seq_len]
        - tail_logtis: [batch_size, predicate_type, seq_len, seq_len]
        返回:
        - batch_size大小的列表，每个元素是一个集合，集合中的元素是三元组
        """
        if threshold is None:
            threshold = self.hparams.threshold
        ent_logits = ent_logits.chunk(ent_logits.shape[0])
        head_logits = head_logits.chunk(head_logits.shape[0])
        tail_logtis = tail_logtis.chunk(tail_logtis.shape[0])
        assert len(head_logits) == len(tail_logtis) == len(ent_logits)
        batch_rels = []
        for i in range(len(ent_logits)):
            subjects, objects = set(), set()
            for l, h, t in zip(*torch.where(ent_logits[i].squeeze(0) > threshold)):
                combined_label = self.hparams.id2combined[l.item()]
                if combined_label[0] == '主体':
                    subjects.add((h, t, combined_label[1])) 
                else:
                    objects.add((h, t, combined_label[1]))
            
            rels = set()
            for sh, st, sl, in subjects:
                for oh, ot, ol in objects:
                    p1s = torch.where(head_logits[i].squeeze(0)[:, sh, oh] > threshold)[0].tolist()
                    p2s = torch.where(tail_logtis[i].squeeze(0)[:, st, ot] > threshold)[0].tolist()
                    ps = set(p1s) & set(p2s)
                    if len(ps) > 0:
                        for p in ps:
                            if mapping is not None:
                                sub_offset = align_token_span((sh.item(), st.item()+1), mapping)
                                obj_offset = align_token_span((oh.item(), ot.item()+1), mapping)
                            else:
                                sub_offset = (sh.item(), st.item()+1)
                                obj_offset = (oh.item(), ot.item()+1)
                            try:
                                sub = Entity(label=sl, indices=[i for i in range(sub_offset[0], sub_offset[1])])
                                obj = Entity(label=ol, indices=[i for i in range(obj_offset[0], obj_offset[1])])
                                rels.add(Relation(s=sub, o=obj, p=self.hparams.id2rel[p]))
                            except:
                                pass
            batch_rels.append(rels)
        return batch_rels
            
        
    def predict(self, text: str, device:str='cpu', threshold = None) -> List[Relation]:
        """模型预测
        参数:
        - text: 要预测的单条文本
        - device: 设备
        - threshold: 三元组抽取阈值, 如果为None, 则为模型训练时的阈值
        返回
        - 预测的三元组
        """
        max_length = min(len(text), self.hparams.plm_max_length)
        inputs = self.tokenizer(
                text, 
                padding='max_length',  
                max_length=max_length,
                return_tensors='pt',
                return_token_type_ids=False,
                truncation=True,
                return_offsets_mapping=True)
        mapping = inputs.pop('offset_mapping')
        mapping = mapping[0].tolist()
        inputs.to(torch.device(device))
        ent_logits, head_logits, tail_logits = self(**inputs)
        if threshold == None:
            batch_rels: List[Set[Relation]] = self.extract_rels(ent_logits, head_logits, tail_logits, threshold=self.hparams.threshold, mapping=mapping)
        else:
            batch_rels: List[Set[Relation]] = self.extract_rels(ent_logits, head_logits, tail_logits, threshold=threshold, mapping=mapping)
        for rel in batch_rels[0]:
            rel.s.text = text[rel.s.indices[0]:rel.s.indices[-1]+1]
            rel.o.text = text[rel.o.indices[0]:rel.o.indices[-1]+1]
        return list(batch_rels[0])


import torch
import torch.nn as nn
import math


def get_sinusoid_encoding_table(n_position, d_hid, padding_idx=None):
    '''Returns: [seq_len, d_hid]
    '''
    table = torch.zeros(n_position, d_hid)
    position = torch.arange(0, n_position, dtype=torch.float).unsqueeze(-1)
    sin_term = torch.pow(10000, -1 * torch.arange(0, d_hid, 2) / d_hid)
    cons_term = torch.pow(10000, -1 * torch.arange(1, d_hid, 2) / d_hid)
    table[..., ::2] = torch.sin(position * sin_term)
    table[..., 1::2] = torch.cos(position * cons_term)
    return table

    #其他实现方式1
    # embeddings_table = torch.zeros(n_position, d_hid)
    # position = torch.arange(0, n_position, dtype=torch.float).unsqueeze(1)
    # sin_div_term = torch.exp(torch.arange(0, d_hid, 2).float() * (-math.log(10000.0) / d_hid))
    # cos_div_term = torch.exp(torch.arange(1, d_hid, 2).float() * (-math.log(10000.0) / d_hid))
    # embeddings_table[:, 0::2] = torch.sin(position * sin_div_term)
    # embeddings_table[:, 1::2] = torch.cos(position * cos_div_term)
    # return embeddings_table

    # 其他实现方式2
    # position_ids = torch.arange(0, n_position).unsqueeze(1)
    # position_ids = position_ids.expand(-1, d_hid)
    # indices = torch.arange(0, d_hid)
    # position_ids = position_ids * torch.pow(10000, -2 * torch.true_divide(torch.floor_divide(indices, 2), d_hid))
    # position_ids[:, ::2] = torch.sin(position_ids[:, ::2])
    # position_ids[:, 1::2] = torch.cos(position_ids[:, 1::2])
    # return position_ids

class RelativePositionEmbedding(nn.Module):
    def __init__(self,
                 qlen,
                 klen,
                 embedding_size,
                 max_relative_position=127) -> None:
        super(RelativePositionEmbedding, self).__init__()
        vocab_size = max_relative_position * 2 + 1
        distance_mat = torch.arange(klen)[None, :] - torch.arange(qlen)[:, None]  # 列数-行数, [query_len, key_len]
        distance_mat_clipped = torch.clamp(distance_mat, -max_relative_position, max_relative_position)
        final_mat = distance_mat_clipped + max_relative_position

        # sinusoid_encoding编码的位置矩阵
        embeddings_table = get_sinusoid_encoding_table(vocab_size, embedding_size)

        # 实现方式1
        # flat_relative_positions_matrix = final_mat.view(-1)
        # one_hot_relative_positions_matrix = torch.nn.functional.one_hot(flat_relative_positions_matrix, num_classes=vocab_size).float()
        # position_embeddings = torch.matmul(one_hot_relative_positions_matrix, embeddings_table)
        # my_shape = list(final_mat.size())
        # my_shape.append(embedding_size)
        # position_embeddings = position_embeddings.view(my_shape)

        # 实现方式2
        # position_embeddings = torch.take_along_dim(embeddings_table, final_mat.flatten().unsqueeze(1), dim=0)
        # position_embeddings = position_embeddings.reshape(*final_mat.shape, embeddings_table.shape[-1])  # [seq_len, seq_len, hdsz]
        # self.register_buffer('position_embeddings', position_embeddings)
        
        # 实现方式3
        position_embeddings = nn.Embedding.from_pretrained(embeddings_table, freeze=True)(final_mat)
        self.register_buffer('position_embeddings', position_embeddings)

    def forward(self, qlen, klen):
        return self.position_embeddings[:qlen, :klen, :]
    
    
# class SinusoidalPositionEmbedding(nn.Module):
#     """定义Sin-Cos位置Embedding
#     """
#     def __init__(self, max_position, embedding_size):
#         super(SinusoidalPositionEmbedding, self).__init__()
#         position_embeddings = nn.Embedding.from_pretrained(get_sinusoid_encoding_table(max_position, embedding_size), freeze=True)
#         self.register_buffer('position_embeddings', position_embeddings)

#     def forward(self, position_ids):
#         return self.position_embeddings(position_ids)


class SinusoidalPositionEmbedding(nn.Module):
    """定义Sin-Cos位置Embedding
    """
    def __init__(
        self,
        output_size,
        merge_mode='add',
        custom_position_ids=False
    ):
        super(SinusoidalPositionEmbedding, self).__init__()
        self.output_size = output_size
        self.merge_mode = merge_mode
        self.custom_position_ids = custom_position_ids

    def forward(self, inputs):
        input_shape = inputs.shape
        _, seq_len = input_shape[0], input_shape[1]
        position_ids = torch.arange(seq_len).type(torch.float)[None]
        indices = torch.arange(self.output_size // 2).type(torch.float)
        indices = torch.pow(10000.0, -2 * indices / self.output_size)
        embeddings = torch.einsum('bn,d->bnd', position_ids, indices)
        embeddings = torch.stack([torch.sin(embeddings), torch.cos(embeddings)], dim=-1)
        embeddings = torch.reshape(embeddings, (-1, seq_len, self.output_size))

        if self.merge_mode == 'add':
            return inputs + embeddings.to(inputs.device)
        elif self.merge_mode == 'mul':
            return inputs * (embeddings + 1.0).to(inputs.device)
        elif self.merge_mode == 'zero':
            return embeddings.to(inputs.device)
        
        
class RoPEPositionEncoding(nn.Module):
    """旋转式位置编码: https://kexue.fm/archives/8265
    """
    def __init__(self, max_position, embedding_size):
        super(RoPEPositionEncoding, self).__init__()
        position_embeddings = get_sinusoid_encoding_table(max_position, embedding_size)  # [seq_len, hdsz]
        cos_position = position_embeddings[:, 1::2].repeat_interleave(2, dim=-1)
        sin_position = position_embeddings[:, ::2].repeat_interleave(2, dim=-1)
        # register_buffer是为了最外层model.to(device)，不用内部指定device
        self.register_buffer('cos_position', cos_position)
        self.register_buffer('sin_position', sin_position)
    
    def forward(self, qw, seq_dim=-2):
        # 默认最后两个维度为[seq_len, hdsz]
        seq_len = qw.shape[seq_dim]
        qw2 = torch.stack([-qw[..., 1::2], qw[..., ::2]], dim=-1).reshape_as(qw)
        return qw * self.cos_position[:seq_len] + qw2 * self.sin_position[:seq_len]


from pydantic import BaseModel, conint, conint, constr, validator, conlist, validate_arguments, conset
from typing import List, Optional, Union, Tuple, DefaultDict, Dict, Any, Set, Generator
import srsly
from pathlib import Path
import pandas as pd
from .dataset import Dataset
from ..utils.text import split_sentence
from tqdm import tqdm
from functools import reduce


Label = constr(strip_whitespace=True, min_length=1)
Index = conint(ge=0, strict=True)


def assert_span_text_in_doc(doc_text: str, span_text: str, span_indices: List[Index]) -> None:
    """检查span的文本与标注的下标对应doc文本一致

    Args:
        doc_text (str): doc文本
        span_text (str): span文本
        span_indices (List[Index]): 在文档中的下标
    """
    try:
        text = ''.join([doc_text[i] for i in span_indices])
    except Exception as e:
        print(span_indices)
        print(len(span_indices))
        raise e
    # 如果实体文本存在则确保实体文本与下标对应文本一致,确保标注正确
    assert text == span_text, f'文本: <{span_text}> 与下标文本: <{text}> 不一致'
    
    
def strip_span(span_text: str, span_indices: List[Index]) -> Tuple[str, List[Index]]:
    """去除文本左右空格字符并将下标对齐

    Args:
        span_text (str): 文本
        span_indices (List[Index]): 下标

    Returns:
        Tuple[str, List[Index]]: 清洗后文本, 对齐后下标
    """
    ori_text = span_text
    text = span_text.strip()
    start = ori_text.index(text)
    indices = span_indices[start: start + len(text)]
    return text, indices
        
    
class Span(BaseModel):
    """原始文本的一个片段,可以连续也可以非连续
    
    参数:
    - text (str): 文本,默认None
    - indices (List[int]): 对应文本的下标
    
    说明:
    - 初始化时如果文本不为空,则会自动去除首尾的空格,
    - 下标会自动按照升序排列
    - 当文本去除首尾空格后,下标会自动修正,例如当text=' 中国'变为'中国', 下标[0,1,2]会变为[1,2]
    """
    text: Optional[constr(min_length=1)] = None
    indices: conset(item_type=int, min_items=1)
    score: Optional[float] = None
    
    @property
    def is_continuous(self) -> bool:
        diff = []
        for i in range(1, len(self.indices)):
            diff.append((self.indices[i] - self.indices[i-1]) != 1)
        return not sum(diff) > 0
            
    @validator('text')
    def validate_text(cls, v: str, values: dict):
        if v is not None:
            values['ori_text'] = v # 记录原始文本,以此修改下边列表
            return v.strip() # 左右没有空格并且有效字符不为0则返回原文
        else:
            return v
    
    @validator('indices')
    def validate_indices(cls, v: List[Index], values):
        v = sorted(v)
        if 'text' in values:
            if values['text']:
                assert len(values['ori_text']) == len(v), f'下标: <{v}>与原始文本: <{values["ori_text"]}>长度不符'
                start = values['ori_text'].index(values['text'])
                indices = v[start: start+len(values['text'])]
                del values['ori_text']
                return indices
            else:
                return v
        else:
            return v
    
    def __hash__(self):
        return hash(self.text)
    
    def __gt__(self, other: "Span") -> bool:
        return self.indices[-1] > other.indices[0] 
    
    def __lt__(self, other: "Span") -> bool:
        return self.indices[-1] < other.indices[0]
    
    def __len__(self):
        return len(self.indices)
    
    def __add__(self, other: "Span"):
        if self.text is not None and other.text is not None:
            text = self.text + other.text
        else:
            text = None
        indices = self.indices + other.indices
        return Span(text=text, indices=indices)
    
    def __and__(self, other: "Span") -> "Span":
        indices = set(self.indices) & set(other.indices)
        if len(indices) == 0:
            return None
        else:
            indices = sorted(list(indices))
            chars = []
            for idx in indices:
                chars.append(self.text[self.indices.index(idx)])
            text = ''.join(chars)
            return Span(indices=indices, text=text)
        
    def __eq__(self, other: "Span") -> bool:
        if self.is_continuous and other.is_continuous:
            return self.indices[0] == other.indices[0] and self.indices[-1] == other.indices[-1]
        else:
            return self.indices == other.indices
    
    def __contains__(self, item: "Span") -> bool:
        if self.is_continuous and item.is_continuous:
            return self.indices[0] <= item.indices[0] and self.indices[-1] >= item.indices[-1]
        else:
            return sorted(self.indices)[0] <= sorted(item.indices)[0] and sorted(self.indices)[-1] >= sorted(item.indices)[-1]

        
class Entity(Span):
    """文本中有标签的Span,支持连续非连续类型
    参数:
    - text (str): 实体文本
    - label (str): 实体标签
    - indices (List[int]): 字符级别下标
    """
    label: Optional[Label] = None
        
    def __hash__(self):
        return hash(self.label)
    
    def __eq__(self, other: "Entity") -> bool:
        if self.is_continuous and other.is_continuous:
            return self.indices[0] == other.indices[0] and self.indices[-1] == other.indices[-1] and self.label == other.label
        else:
            return self.indices == other.indices and self.label == other.label
    
    
class Relation(BaseModel):
    """实体之间的有指向关系

    参数:
    - s(Entity): 主体
    - p(str): 关系标签
    - o(Entity): 客体
    """
    s: Entity
    p: Label
    o: Entity
    
    @validator('o')
    def validate_rel(cls, v, values):
        assert values['s'] != v, '主体客体同一实体'
        return v
    
    def __hash__(self):
        return hash(self.p)
    
    def __eq__(self, other: "Relation") -> bool:
        return self.s == other.s and self.p == other.p and self.o == other.o
    
    
class Event(BaseModel):
    """由若干实体和触发词组成的事件
    - args (List[Entity]): 组成事件的论元
    - label (Label): 事件类型标签
    - trigger(Optional[Span]): 触发词span,默认为None
    """
    
    args: conlist(item_type=Entity, min_items=1)
    label: Label
    trigger: Optional[Span] = None
    
    @validator('args')
    def validate_args(cls, v):
        return list(set(v))
    
    @validate_arguments
    def add_arg(self, arg: Entity):
        if arg not in self.args:
            self.args.append(arg)
    
    def __hash__(self):
        return hash(self.label)
    
    def __eq__(self, other: "Event") -> bool:
        return self.args == other.args and self.label == other.label and self.trigger == other.trigger
    
    
class Answer(BaseModel):
    spans: List[Span] = []
    
    @property
    def text(self):
        if len(self.spans) == 0:
            return ''
        else:
            text_list = []
            for span in self.spans:
                if span.text not in text_list and span.text is not None:
                    text_list.append(span.text)
            return ' '.join(text_list)
        
    def add_span(self, span: Span):
        if span not in self.spans:
            self.spans.append(span)
        self.spans = sorted(self.spans)


class Doc(BaseModel):
    """存放所有标注数据,由模型或者原始数据构建
    参数:
    - text(str): 文本
    - id(str): 唯一标识符
    - label(str): 唯一标签
    - labels(List[str]): 多标签
    - ents(List[Entity]): 实体列表
    - rels(List[Relation]): 关系列表
    - events(List[Event]): 事件列表
    - sents(List(Span)): 文本连续的句子,不需要初始化,自动生成
    - summary(str): 摘要
    - title(str): 标题
    - questions(Dict[str, Answer]): 问题对应答案
    """
    
    text: constr(min_length=1)
    id: Optional[str] = None
    label: Optional[Label] = None
    labels: Optional[List[Label]] = None
    ents: Optional[List[Entity]] = None
    rels: Optional[List[Relation]] = None
    events: Optional[List[Event]] = None   
    summary: Optional[str] = None
    title: Optional[str] = None
    questions : Optional[Dict[str, Answer]] = None
    
    @property
    def sents(self) -> Generator:
        start = 0
        for s in split_sentence(self.text, best=False):
            end = start + len(s)
            yield Span(text=s, indices=[i for i in range(start, end)])
            start = end
    
    @validator('text')
    def validate_text(cls, v: str):
        assert len(v.strip()) > 0, f'文本为空格'
        return v
    
    @validator('questions')
    def validate_questions(cls, v: Dict, values):
        for q, a in v.items():
            a: Answer
            for span in a.spans:
                assert type(span) == Span, '答案必须为Span'
                if span.text:
                    assert_span_text_in_doc(doc_text=values['text'], span_indices=span.indices, span_text=span.text)
        return v
    
    @validator('ents', each_item=True)
    def validate_ents(cls, v: Entity, values):      
        if v.text:
            assert_span_text_in_doc(doc_text=values['text'], span_indices=v.indices, span_text=v.text)          
        return v
    
    @validator('rels', each_item=True)
    def validate_rels(cls, v: Relation, values):
        text = values['text']
        if v.s.text:
            assert_span_text_in_doc(doc_text=text, span_text=v.s.text, span_indices=v.s.indices)
        if v.o.text:
            assert_span_text_in_doc(doc_text=text, span_text=v.o.text, span_indices=v.o.indices)
        return v
    
    @validator('events', each_item=True)
    def validate_events(cls, v: Event, values):
        text = values['text']
        for arg in v.args:
            arg: Entity
            if arg.text:
                assert_span_text_in_doc(doc_text=text, span_text=arg.text, span_indices=arg.indices) 
        if v.trigger:
            trigger : Span = v.trigger
            if trigger.text:
                assert_span_text_in_doc(doc_text=text, span_indices=trigger.indices, span_text=trigger.text)
        return v
    
    def _get_indices_text(self, indices: List[Index]) -> str:
        return ''.join([self.text[i] for i in indices])
    
    @validate_arguments
    def add_ent(self, indices: List[int], label: str, text: Optional[str] = None, score: Optional[float] = None):
        if text:
            assert_span_text_in_doc(doc_text=self.text, span_indices=indices, span_text=text)
        else:
            text = self._get_indices_text(indices=indices)
            text, indices = strip_span(span_text=text, span_indices=indices)
        ent = Entity(text=text, indices=indices, label=label, score=score)
        if not self.ents:
            self.ents = [ent]
        else:
            if ent not in self.ents:
                self.ents.append(ent)
            
    @validate_arguments
    def add_rel(self, rel: Relation):
        if rel.s.text:
            assert_span_text_in_doc(doc_text=self.text, span_indices=rel.s.indices, span_text=rel.s.text)
        else:
            s_text = self._get_indices_text(indices=rel.s.indices)
            s_text, s_indices = strip_span(span_text=s_text, span_indices=rel.s.indices)
            rel.s.text = s_text
            rel.s.indices = s_indices
        if rel.o.text:
            assert_span_text_in_doc(doc_text=self.text, span_indices=rel.o.indices, span_text=rel.o.text)
        else:
            o_text = self._get_indices_text(indices=rel.o.indices)
            o_text, o_indices = strip_span(span_text=o_text, span_indices=rel.o.indices)
            rel.o.text = o_text
            rel.o.indices = o_indices
        if not self.rels:
            self.rels = [rel]
        else:
            if rel not in self.rels:
                self.rels.append(rel)

    @validate_arguments
    def add_event(self, event: Event):
        for arg in event.args:
            arg: Entity
            if arg.text:
                assert_span_text_in_doc(doc_text=self.text, span_indices=arg.indices, span_text=arg.text)
            else:
                arg_text = self._get_indices_text(indices=arg.indices)
                arg_text, arg_indices = strip_span(span_text=arg_text, span_indices=arg.indices)
                arg.text = arg_text
                arg.indices = arg_indices
        if event.trigger:
            if event.trigger.text:
                assert_span_text_in_doc(doc_text=self.text, span_indices=event.trigger.indices, span_text=event.trigger.text)
            else:
                trigger_text = self._get_indices_text(indices=event.trigger.indices)
                trigger_text, trigger_indices = strip_span(span_text=trigger_text, span_indices=event.trigger.indices)
                event.trigger.text = trigger_text
                event.trigger.indices = trigger_indices
        if not self.events:
            self.events = [event]
        else:
            if event not in self.events:
                self.events.append(event)
            
    @validate_arguments
    def add_label(self, label: Label):
        if not self.labels:
            self.labels = [label]
        else:
            if label not in self.labels:
                self.labels.append(label)
    
    @validate_arguments
    def set_label(self, label: Label):
        self.label = label
        
    @validate_arguments
    def set_summary(self, summary: constr(strip_whitespace=True, strict=True, max_length=1)):
        self.summary = summary
        
    @validate_arguments
    def set_title(self, title: constr(strip_whitespace=True, strict=True, max_length=1)):
        self.title = title
        
    @validate_arguments
    def add_question(self, question: constr(min_length=1, strip_whitespace=True)) -> None:
        if self.questions is None:
            self.questions = {}  
        if not question in self.questions:
            empty_ans = Answer()
            self.questions[question] = empty_ans
            
    def _combine_spans(self, spans: List[Span]) -> Optional[Span]:
        if len(spans) == 0:
            return None
        else:
            indices = [span.indices for span in spans]
            indices = reduce(add, indices)
            indices = sorted(list(set(indices)))
            text = self._get_indices_text(indices=indices)
            return Span(text=text, indices=indices)
        
    def combine_spans_to_contiguous(self, spans: List[Span]) -> Optional[Span]:
        if len(spans) == 0:
            return None
        else:
            start = spans[0].indices[0]
            end = spans[-1].indices[-1] + 1
            indices = [i for i in range(start, end)]
            text = self._get_indices_text(indices=indices)
            return Span(text=text, indices=indices)

    def split_by_sents(self, max_length: int) -> List[Span]:
        """将文本按照句子切分为不超过固定长度的片段

        Args:
            max_piece_length (int): 最大片段长度

        Returns:
            List[span]: 切分后的片段列表
        """
        pieces = []
        piece: List[Span] = []
        for sent in self.sents:
            sent: Span
            if len(piece)>0:
                cur_length = sent.indices[-1] + 1 - piece[0].indices[0]
            else:
                cur_length = sent.indices[-1] + 1
            if cur_length < max_length:
                piece.append(sent)
            else:
                if len(piece) > 0:
                    pieces.append(self.combine_spans_to_contiguous(spans=piece))
                if len(sent) > max_length:
                    continue
                else:
                    piece = [sent]
        if len(piece) > 0:
            pieces.append(self.combine_spans_to_contiguous(spans=piece))
        return pieces
    
    def get_answer(self, question: str) -> Answer:
        q = question.strip()
        return self.questions.get(q)
    
    def add_answer_span(self, question: str, answer_indices: List[int], answer_text: Optional[str] = None) -> None:
        question = question.strip()
        ans = self.questions.get(question)
        assert ans is not None, f'未能找到相应问题:{question},请确保问题已经设置'
        if answer_text is None:
            answer_text = self._get_indices_text(indices=answer_indices)
        else:
            assert_span_text_in_doc(doc_text=self.text, span_text=answer_text, span_indices=answer_indices)
        span = Span(indices=answer_indices, text=answer_text)
        ans.add_span(span=span)
        
        
    class Config:
        extra = 'forbid'
        allow_mutation = True
        
    def __hash__(self):
        return hash(self.text)
    
    def __eq__(self, other: "Doc") -> bool:
        if self.id and other.id:
            return self.id == other.id
        else:
            return self.text == other.text and \
                   self.label == other.label and \
                   self.labels == other.labels and \
                   self.ents == other.ents and \
                   self.rels == other.rels and \
                   self.events == other.events and \
                   self.summary == other.summary 
                
                   
class DocBin():
    """存放所有doc
    参数:
    - docs (Union[List[Doc], Path]): 所有Doc文档示例或者jsonl格式地址
    """
    def __init__(self, docs : Union[List[Doc], Path] = None) -> None:
        super().__init__()
        if type(docs) == str:
            self._docs = self._get_docs_from_jsonl(docs)
        elif docs is None:
            self._docs = []
        else:
            self._docs = docs
            
    def __getitem__(self, i):
        return self._docs[i]
    
    def __len__(self):
        return len(self._docs)
    
    def __repr__(self) -> str:
        return f"{len(self._docs)} docs"
    
    def __str__(self) -> str:
        return f"{len(self._docs)} docs"
    
    def __add__(self, other: Union["DocBin", List[Doc]]) -> "DocBin":
        if isinstance(other, list):
            return DocBin(self._docs + other)
        if isinstance(other, DocBin):
            return DocBin(self._docs + other._docs)
        
    def save_to_disk(self, file_path: Path):
        """将数据以jsonl的格式保存到硬盘
        参数:
        - file_path (Path): 数据保存地址,例如./test.jsonl
        """
        path = Path(file_path)
        srsly.write_jsonl(path=path, lines=[doc.dict() for doc in self._docs])
    
    def _get_docs_from_jsonl(self, file_path: Path) -> List[Doc]:
        path = Path(file_path)
        datas = srsly.read_jsonl(path=path)
        docs = []     
        for d in datas:
            docs.append(Doc(**d))
        return docs
        
    def to_dataset(self, include: Optional[List] = None) -> Dataset:
        """转换数据集,None的数据自动去除
        参数:
        - include (List): 包含的字段名称,默认None
        """
        if include:
            datas = [doc.dict(include=set(include), exclude_none=True) for doc in self._docs]
        else:
            datas = [doc.dict(exclude_none=True) for doc in self._docs]
        return Dataset.from_pandas(pd.DataFrame.from_dict(datas))
    
    def to_dataframe(self, include: Optional[List] = None, dropna: bool = False) -> pd.DataFrame:
        """转换为dataframe格式
        """
        data = [doc.dict() for doc in self._docs]
        df = pd.DataFrame.from_records(data)
        if include:
            df = df.loc[:, include]
            if dropna:
                return df.dropna(how='all', axis=1)
            else:
                return df
        else:
            if dropna:
                return df.dropna(how='all', axis=1)
            else:
                return df
            
    @classmethod
    def from_pandas(cls, df: pd.DataFrame) -> "DocBin":
        records = df.to_dict(orient='records')
        docs = [Doc(**r) for r in records]
        return DocBin(docs=docs)
    
    def to_ner_dataset(self, piece_max_length: int = 500, only_have_ent: bool = True) -> Dataset:
        """转换为实体抽取数据集
        - piece_max_length (int): 长文本切分的每个片段长度
        - only_have_ent (bool): 是否仅保存含有实体的数据
        """
        data = {'text': [], 'ents': []}
        for doc in self._docs:
            doc: Doc
            for piece in doc.split_by_sents(max_length=piece_max_length):
                ents = []
                for ent in doc.ents:
                    if ent in piece:
                        ents.append(ent.dict())
                if len(ents)>0:
                    data['text'].append(piece.text)
                    data['ents'].append(ents)
        return Dataset.from_dict(data)
    
    def to_tc_dataset(self) -> Dataset:
        """转换为单目标文本分类数据集
        """
        df = self.to_dataframe(include=['text', 'label'])
        df = df[df['label'].notna()]
        assert len(df)>0, '数据集为空'
        return Dataset.from_pandas(df, preserve_index=False)
    
    def to_re_dataset(self) -> Dataset:
        """转换为实体关系抽取数据集
        """
        df = self.to_dataframe(include=['text', 'rels'])
        df = df[df['rels'].notna()]
        assert len(df)>0, '数据集为空'
        return Dataset.from_pandas(df, preserve_index=False)
    
    def to_ee_dataset(self) -> Dataset:
        """转换为事件抽取数据集
        """
        df = self.to_dataframe(include=['text', 'events'])
        df = df[df['events'].notna()]
        assert len(df)>0, '数据集为空'
        return Dataset.from_pandas(df, preserve_index=False)
    
    def to_summary_dataset(self) -> Dataset:
        """转换为文本摘要数据集
        """
        df = self.to_dataframe(include=['text', 'summary'])
        df = df[df['summary'].notna()]
        assert len(df)>0, '数据集为空'
        return Dataset.from_pandas(df, preserve_index=False)
    
    def to_qa_dataset(self, max_length: int = 450, only_have_answer: bool = False) -> Dataset:
        """转换为问答数据集

        Args:
            max_length (int, optional): 按照句子级别切分的最大文本长度. Defaults to 450.
            only_have_answer (bool): 仅选取有答案的数据. Defaults to False.

        Returns:
            Dataset : 按照句子切分后的问答数据集
        """
        data = {'text': [], 'question': [], 'answer': [], 'spans': []}
        for doc in tqdm(self._docs):
            doc: Doc
            pieces = doc.split_by_sents(max_length=max_length)
            for p in pieces:
                p: Span
                for q, a in doc.questions.items():
                    a: Answer
                    new_a = Answer()
                    for span in a.spans:
                        _span = span & p
                        if _span:
                            indices = [i - p.indices[0] for i in _span.indices]
                            p_span = Span(text=_span.text, indices=indices)
                            assert_span_text_in_doc(doc_text=p.text, span_text=p_span.text, span_indices=p_span.indices)
                            new_a.add_span(span=p_span)
                    if only_have_answer:
                        if len(new_a.spans) > 0:
                            data['text'].append(p.text)
                            data['question'].append(q)
                            data['answer'].append(new_a.text)
                            data['spans'].append(new_a.dict()['spans'])  
                    else:
                        data['text'].append(p.text)
                        data['question'].append(q)
                        data['answer'].append(new_a.text)
                        data['spans'].append(new_a.dict()['spans'])
        return Dataset.from_dict(data)
    
    @validate_arguments
    def append(self, doc: Doc):
        self._docs.append(doc)
    
    @validate_arguments
    def add(self, doc: Doc):
        if doc not in self._docs:
            self._docs.append(doc)


from ..utils.make_datamodule import char_idx_to_token, PLMBaseDataModule
import torch


class PromptSpanExtractionDataModule(PLMBaseDataModule):
    """
    Data module for the prompt span extraction task.

    Dataset examples:
    {'text':'北京是中国的首都', 'prompt':'国家', 'spans':[{'text': '中国', 'offset':(3, 5)}]}
    {'text':'北京是中国的首都', 'prompt':'中国的首都', 'spans':[{'text': '北京', 'offset':(0, 2)}]}
    {'text': '北京是中国的首都', 'prompt': '北京的国家', 'spans': [{'text': '中国', 'offset': (3, 5)}]}
    其中offset 为左闭右开的字符级别下标
    """
    
    def __init__(self,
                 dataset: str,
                 plm: str,
                 transform: str,
                 batch_size: int,
                 max_length: int = -1,
                 pin_memory: bool = False,
                 num_workers: int = 0,
                 dataset_dir: str = 'datasets',
                 plm_dir: str = 'plms') :
        """基于模板提示的文本片段抽取数据模块

        Args:
            dataset (str): 数据集名称
            plm (str): 预训练模型名称
            max_length (int): 文本最大长度
            batch_size (int): 批次大小
        """
        super().__init__()
        
        self.transforms = {'global_span': self.global_span_transform}
        
        
    def setup(self, stage: str) -> None:
        self.dataset.set_transform(transform=self.transforms.get(self.hparams.transform))
        
 
    def global_span_transform(self, example):
        batch_text = example['text']
        batch_spans = example['spans']
        batch_prompt = example['prompt']
        max_length = self.hparams.max_length
        batch = {'inputs': [], 'span_ids': []}
        for i, text in enumerate(batch_text):
            # tokens = fine_grade_tokenize(text, self.tokenizer)
            prompt = batch_prompt[i]
            inputs = self.tokenizer(
                prompt, 
                text, 
                padding='max_length',  
                max_length=max_length,
                truncation=True,
                return_offsets_mapping=True)
            offset_mapping = [list(x) for x in inputs["offset_mapping"]]
            bias = 0
            for index in range(len(offset_mapping)):
                if index == 0:
                    continue
                mapping = offset_mapping[index]
                if mapping[0] == 0 and mapping[1] == 0 and bias == 0:
                    bias = index
                if mapping[0] == 0 and mapping[1] == 0:
                    continue
                offset_mapping[index][0] += bias
                offset_mapping[index][1] += bias
            spans = batch_spans[i]
            span_ids = torch.zeros(1, max_length, max_length)
            for span in spans:
                start = char_idx_to_token(span['offset'][0]+bias, offset_mapping)
                end = char_idx_to_token(span['offset'][1]-1+bias, offset_mapping)
                # 如果超出边界，则跳过
                try:
                    span_ids[0, start, end] = 1.0
                except Exception:
                    self.log.warning(f'set span {(start, end)} out of boudry')
                    pass 
            del inputs['offset_mapping']
            inputs = dict(zip(inputs.keys(), map(torch.tensor, inputs.values())))
            batch['inputs'].append(inputs)
            batch['span_ids'].append(span_ids)
        return batch


import torch.nn as nn

class SimpleDense(nn.Module):
    def __init__(self,
                input_size: int,
                hidden_size: int,
                output_size: int):
        super(SimpleDense, self).__init__()

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.lin1 = nn.Linear(self.input_size, self.hidden_size)
        self.lin2 = nn.Linear(self.hidden_size, self.output_size)
        self.layer_norm = nn.LayerNorm(self.hidden_size)
        self.activation = nn.ReLU()

        
    def forward(self, x):
        x = self.lin1(x)
        x = self.layer_norm(x)
        x = self.activation(x)
        x = self.lin2(x)
        return x


from lightning.pytorch import Callback
import torch
import os
import logging

log = logging.getLogger(__name__)


class LoadModelStateDict(Callback):
    def __init__(self, ckpt_path: str) -> None:
        super().__init__()
        self.ckpt_path = ckpt_path
        
    def on_train_start(self, trainer, pl_module) -> None:
        state_dict = torch.load(self.ckpt_path)['state_dict']
        log.info((f'Loading state from {self.ckpt_path} on device {pl_module.device}'))
        pl_module.load_state_dict(state_dict)
        log.info('all weights loaded')
        

class LoadPLMStateDict(Callback):
    """训练开始前加载预训练模型
    
    - 预训练模型的名称必须为plm bert encoder decoder 之一
    
    """

    def on_train_start(self, trainer, pl_module):
        ckpt_path = os.path.join(pl_module.hparams.plm_dir, pl_module.hparams.plm, 'pytorch_model.bin')
        log.info((f'Loading PLM state dict from {ckpt_path} on device {pl_module.device}'))
        try: 
            pl_module.plm.load_state_dict(torch.load(ckpt_path))
        except Exception:
            pass
        try: 
            pl_module.bert.load_state_dict(torch.load(ckpt_path))
        except Exception:
            pass
        try: 
            pl_module.encoder.load_state_dict(torch.load(ckpt_path))
        except Exception:
            pass
        try:
            pl_module.decoder.load_state_dict(torch.load(ckpt_path))
        except Exception:
            pass
        log.info('all weights loaded')


from ..utils.make_datamodule import PLMBaseDataModule, char_idx_to_token
import torch
from functools import lru_cache


class SpanExtractionDataModule(PLMBaseDataModule):
    """span分类的数据模块 数据集必须有text, spans两个字段
    """
    def __init__(self,
                dataset: str,
                plm: str,
                batch_size: int,
                transform: str = 'globalpointer',
                auto_length: str='max',
                **kwargs):
        super().__init__()
        self.transforms['globalpointer'] = self.gp_transform

    @property
    @lru_cache()
    def label2id(self):
        set_labels = sorted(set([span['label'] for spans in self.dataset['train']['spans'] for span in spans]))
        label2id = {label: i for i, label in enumerate(set_labels)}
        return label2id


    def gp_transform(self, example):
        batch_text = example['text']
        batch_spans = example['spans']
        max_length = self.hparams.max_length
        batch_inputs = {'input_ids':[],'token_type_ids':[],'attention_mask':[]}
        batch_span_ids = []
        for i, text in enumerate(batch_text):
            inputs = self.tokenizer(text, 
                                    padding='max_length',  
                                    max_length=max_length,
                                    truncation=True,
                                    return_offsets_mapping=True)
            spans = batch_spans[i]
            mapping = inputs['offset_mapping']
            span_ids = torch.zeros(len(self.hparams.label2id), max_length, max_length)
            for span in spans :
                # +1 是因为添加了 [CLS]
                start = span['offset'][0]
                start = char_idx_to_token(start, mapping)
                end = span['offset'][1] 
                end = char_idx_to_token(end, mapping)
                label_id = self.hparams.label2id[span['label']]
                span_ids[label_id,  start, end] = 1
            batch_inputs['input_ids'].append(inputs['input_ids'])
            batch_inputs['token_type_ids'].append(inputs['token_type_ids'])
            batch_inputs['attention_mask'].append(inputs['attention_mask'])
            batch_span_ids.append(span_ids)
        batch_span_ids = torch.stack(batch_span_ids, dim=0)
        batch = dict(zip(batch_inputs.keys(), map(torch.tensor, batch_inputs.values())))
        batch['span_ids'] = batch_span_ids
        return batch


    def setup(self, stage: str) -> None:
        self.hparams.max_length = self.get_max_length()
        self.hparams.label2id = self.label2id
        self.dataset.set_transform(transform=self.transforms.get(self.hparams.transform))

    @staticmethod
    def get_one_sample(self):
        return {'text':'我的电话是12345','spans':[{'label':'电话','offset':[5, 10], 'text':'12345'}]}


from datasets import Dataset, DatasetDict
from typing import Dict, List, Optional, Union, Tuple
from ..algorithms.text_match import BM25
import random
from tqdm import tqdm

        
def make_text_match_dataset_with_bm25(corpus: List[str],
                                     synonym_dict: Optional[Dict[str, set]] = None,
                                     num_positive_samples: int=10,
                                     num_negative_samples: int=20,
                                     recall_topk: int = 1000,
                                     reverse_sample: bool = False,
                                     positive_label: str = '1',
                                     negative_label: str='0',
                                     tokenizer = None,
                                     k1: float = 1.5,
                                     b: float=0.75,
                                     epsilon: float=0.25,
                                     is_retrain_docs=True,
                                     return_bm25: bool = False):
    """制作文本匹配二分类数据集, 此数据集适用于text_pair_classification任务

    Args:
        corpus (List[str]): 所有的可以召回的语料,词语等
        synonym_dict (Dict[str, set]): 标准词对应的同义词字典.字典的键应该包含于corpus.
        num_negative_samples (int): 负样本的数量.
        recall_topk(int): 召回模型的召回数量
        reverse_sample (bool): 是否逆序负样本采样
        positive_label (str): 正样本标签, 默认为'1'.
        negative_label (str): 负样本标签, 默认为'0'.
        tokenizer (_type_, optional): 分词器,默认为字符切分. Defaults to None.
        k1 (float, optional): bm25参数. Defaults to 1.5.
        b (float, optional): bm25参数. Defaults to 0.75.
        epsilon (float, optional): bm25参数. Defaults to 0.25.
        is_retrain_docs (bool, optional): bm25参数, 是否保留文档. Defaults to True.
        return_bm25 (bool): 是否返回bm25模型. Defaults to False.
    """
    bm25 = BM25(corpus=corpus, 
                k1=k1, 
                b=b, 
                epsilon=epsilon,
                is_retain_docs=is_retrain_docs,
                tokenizer=tokenizer)
    label_ls = []
    text_a_ls = []
    text_b_ls = []
    for key in tqdm(synonym_dict.keys()):
        recalls = bm25.recall(key, topk=recall_topk)
        recalls = [r[0] for r in recalls if r[0] != key and r[0] not in synonym_dict[key]]
        if reverse_sample:
            recalls.reverse()
            negatives = recalls[:num_negative_samples]
        else:
            negatives = recalls[:num_negative_samples]
        text_a_ls.extend(negatives)
        text_b_ls.extend([key]*len(negatives))
        label_ls.extend([negative_label]*len(negatives))
        syms = list(synonym_dict[key])
        positives = random.choices(syms, k=num_positive_samples)
        text_a_ls.extend(positives)
        text_b_ls.extend([key]*num_positive_samples)
        label_ls.extend([positive_label]*num_positive_samples)
    ds = Dataset.from_dict({'text_a':text_a_ls, 'text_b':text_b_ls, 'label': label_ls})
    if not return_bm25:
        return ds
    else:
        return ds, bm25
    
def train_val_split(dataset:Dataset,
                    val_frac: float =0.1,
                    return_dataset_dict: bool =True) -> Union[Tuple["Dataset", "Dataset"], DatasetDict]:
        """split dataset into tarin and validation datasets
        Args:
            dataset (Dataset): dataset to split
            val_frac (float, optional): validation radio of all dataset. Defaults to 0.1.
            return_dataset_dict (bool, optional): if return_dataset_dict is True, return a DatasetDict,
                otherwise return a tuple of train, val datasets. Defaults to True.

        Returns:
            Union[Tuple[Dataset, Dataset], DatasetDict]: if return_dataset_dict is True, return a DatasetDict, otherwise return a tuple of train, val datasets
        """
        df = dataset.to_pandas()
        train_df = df.sample(frac=1-val_frac)
        val_df = df.drop(train_df.index)
        train_ds = Dataset.from_pandas(train_df, preserve_index=False)
        val_ds = Dataset.from_pandas(val_df, preserve_index=False)
        if not return_dataset_dict:
            return train_ds, val_ds
        else:
            return DatasetDict({'train': train_ds, 'validation': val_ds})
        
    
def train_val_test_split(dataset: Dataset,
                         val_frac: float =0.1,
                         test_frac: float =0.1,
                         return_dataset_dict: bool =True) -> Union[Tuple["Dataset", "Dataset", "Dataset"], DatasetDict]:
    """split dataset into tarin vlidation and test datasets

    Args:
        dataset (Dataset): dataset to split
        val_frac (float, optional): validation radio of all dataset. Defaults to 0.1.
        test_frac (float, optional): test radio of all dataset. Defaults to 0.1.

    Returns:
        Union[Tuple[Dataset, Dataset, Dataset], DatasetDict]: if return_dataset_dict is True, return a DatasetDict, 
            otherwise return a tuple of train, val, test datasets
        
    """
    df = dataset.to_pandas()
    train_df = df.sample(frac=1-val_frac-test_frac)
    other_df = df.drop(train_df.index)
    val_df = other_df.sample(frac=1-(test_frac/(val_frac+test_frac)))
    test_df = other_df.drop(val_df.index)
    train_ds = Dataset.from_pandas(train_df, preserve_index=False)
    val_ds = Dataset.from_pandas(val_df, preserve_index=False)
    test_ds = Dataset.from_pandas(test_df, preserve_index=False)
    if not return_dataset_dict:
        return train_ds, val_ds, test_ds
    else:
        return DatasetDict({'train': train_ds, 'validation': val_ds, 'test': test_ds})


import torch
from torchmetrics import F1Score
from typing import List, Any, Dict
import torch.nn.functional as F
from ...utils.make_model import PLMBaseModel
from transformers import AutoModelForSequenceClassification



class BERTCrossEncoder(PLMBaseModel):
    '''
    交互式的bert句子对分类模型
    '''
    def __init__(self, 
                 lr: float = 3e-5,
                 weight_decay: float = 0.01,
                 scheduler: str = 'linear_warmup',
                 **kwargs):
        super().__init__()  
        
        # 模型结构
        self.plm = self.get_plm_architecture()
        # 损失函数
        self.criterion = torch.nn.CrossEntropyLoss()
        # 指标评价函数
        self.train_f1 = F1Score(num_classes=len(self.hparams.label2id))
        self.val_f1 = F1Score(num_classes=len(self.hparams.label2id))
        self.test_f1 = F1Score(num_classes=len(self.hparams.label2id))
    
    def get_plm_architecture(self) -> torch.nn.Module:
        trf_config = self.trf_config
        trf_config.id2label = self.hparams.id2label
        trf_config.labels = len(self.hparams.label2id)
        return AutoModelForSequenceClassification.from_config(self.trf_config)
            
    def setup(self, stage: str) -> None:
        self.trainer.datamodule.dataset.set_transform(self.trainer.datamodule.cross_transform)     

    def forward(self, input_ids, token_type_ids, attention_mask):
        x = self.plm(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask)
        return x.logits

    def shared_step(self, batch):
        input_ids = batch['input_ids']
        token_type_ids = batch['token_type_ids']
        attention_mask = batch['attention_mask']
        labels = batch['label_ids']
        logits = self(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask)
        loss = self.criterion(logits, labels)
        preds = torch.argmax(logits, dim=-1)
        return loss, preds, labels

    def training_step(self, batch, batch_idx):
        loss, preds, labels = self.shared_step(batch)
        self.train_f1(preds, labels)
        self.log('train/f1', self.train_f1, on_step=True, on_epoch=True, prog_bar=True)
        return {'loss': loss}

    def validation_step(self, batch, batch_idx):
        loss, preds, labels = self.shared_step(batch)
        self.val_f1(preds, labels)
        self.log('val/f1', self.val_f1, on_step=False, on_epoch=True, prog_bar=True)


    def test_step(self, batch, batch_idx):
        loss, preds, labels = self.shared_step(batch)
        self.test_f1(preds, labels)
        self.log('test/f1', self.test_f1, on_step=False, on_epoch=True, prog_bar=True)


    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.plm.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.plm.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0},
        ]
        optimizer = torch.optim.AdamW(grouped_parameters)
        scheduler = self.get_scheduler_config(optimizer=optimizer, name=self.hparams.scheduler)
        return [optimizer], [scheduler]
        
        
    def predict(self, batch_text_a: List[str], batch_text_b: List[str], device: str='cpu') -> Dict[str, float]:
        device = torch.device(device)
        inputs = self.tokenizer(batch_text_a,
                                batch_text_b,
                                padding=True,
                                max_length=512,
                                return_tensors='pt',
                                truncation=True)
        inputs.to(device)
        self.eval()
        with torch.no_grad():
            logits = self(**inputs)
            label_ids = logits.argmax(dim=-1).tolist()
            labels = [self.hparams.id2label[i] for i in label_ids]
        return labels


from functools import lru_cache
from ..utils.make_datamodule import PLMBaseDataModule, char_idx_to_token
from ..utils.utils import get_logger
import pandas as pd
import numpy as np
import torch
from typing import List


log = get_logger()

class EntityExtractionDataModule(PLMBaseDataModule):
    """实体抽取数据模块
    可以解决嵌套实体和非连续实体
    globalpointer: 可以抽取嵌套实体,不能抽取非连续实体
    w2ner: 可以抽取非连续和嵌套实体
    
    数据集格式: {"text":"这是一个长颈鹿","ents":[{"indices":[4,5,6],"label":"动物", "text":"长颈鹿"}]}
    """
        
    def __init__(self,
                 dataset: str,
                 batch_size: int,
                 plm: str = 'hfl/chinese-roberta-wwm-ext',
                 **kwargs):
        super().__init__()
        
    def setup(self, stage: str) -> None:
        self.hparams.id2ent = self.id2ent
        self.hparams.id2bio = self.id2bio
         
    
    @property
    @lru_cache()
    def ent_labels(self) -> List:
        labels = sorted(pd.Series(np.concatenate(self.train_df['ents'])).apply(lambda x: x['label']).drop_duplicates().values)
        return labels
    
    @property
    def ent2id(self):
        return {l:i for i,l in enumerate(self.ent_labels)}

    @property
    def id2ent(self):
        return {i:l for l,i in self.ent2id.items()}
    
    @property
    @lru_cache()
    def bio_labels(self) -> List:
        b_labels = ['B' + '-' + l for l in self.ent_labels]
        i_labels = ['I' + '-' + l for l in self.ent_labels]
        return ['O'] + b_labels + i_labels
    
    @property
    def bio2id(self):
        return {l:i for i,l in enumerate(self.bio_labels)}
    
    @property
    def id2bio(self):
        return {i:l for l,i in self.bio2id.items()}

    @lru_cache()
    def get_dis2idx(self):
        dis2idx = np.zeros((1000), dtype='int64')
        dis2idx[1] = 1
        dis2idx[2:] = 2
        dis2idx[4:] = 3
        dis2idx[8:] = 4
        dis2idx[16:] = 5
        dis2idx[32:] = 6
        dis2idx[64:] = 7
        dis2idx[128:] = 8
        dis2idx[256:] = 9
        return dis2idx


    def fill(self, data, new_data):
        for j, x in enumerate(data):
            new_data[j, :x.shape[0], :x.shape[1]] = np.array(x, dtype=np.int32)
        return new_data


    def w2ner_transform(self, examples):
        max_length = self.get_max_length()
        batch_text = examples['text']
        batch_ents = examples['ents']
        batch_inputs = self.tokenizer(batch_text,
                                      max_length=max_length,
                                      padding='max_length',
                                      truncation=True,
                                      return_offsets_mapping=True,
                                      add_special_tokens=False,
                                      return_tensors='pt')
        batch_mappings = batch_inputs.pop('offset_mapping').tolist()
        batch_lengths = batch_inputs['attention_mask'].sum(dim=-1).tolist()
        batch_size = len(batch_text)
        batch_dist_ids = np.zeros((batch_size, max_length, max_length), dtype=np.int)
        batch_dist = []
        batch_label_ids = []
        dis2idx = self.get_dis2idx()
        for i, text in enumerate(batch_text):
            mapping = batch_mappings[i]
            ents = batch_ents[i]
            length = batch_lengths[i]
            label_ids = np.zeros((max_length, max_length), dtype=np.int)
            for ent in ents:
                idx_ls = ent['indices']
                for i in range(len(idx_ls)):
                    if i +1 >= len(idx_ls):
                        break
                    label_ids[char_idx_to_token(idx_ls[i], mapping), char_idx_to_token(idx_ls[i+1], mapping)] = 1
                label_ids[char_idx_to_token(idx_ls[-1], mapping), char_idx_to_token(idx_ls[0], mapping)] = self.ent2id[ent['label']]
            batch_label_ids.append(label_ids)

            _dist_inputs = np.zeros((length, length), dtype=np.int)
            for k in range(length):
                _dist_inputs[k, :] += k
                _dist_inputs[:, k] -= k

            for i in range(length):
                for j in range(length):
                    if _dist_inputs[i, j] < 0:
                        _dist_inputs[i, j] = dis2idx[-_dist_inputs[i, j]] + 9
                    else:
                        _dist_inputs[i, j] = dis2idx[_dist_inputs[i, j]]
            _dist_inputs[_dist_inputs == 0] = 19
            batch_dist.append(_dist_inputs)

        batch_label_ids = np.stack(batch_label_ids)
        batch_inputs['label_ids'] = torch.from_numpy(batch_label_ids)
        _batch_dist_ids = self.fill(batch_dist, batch_dist_ids)
        batch_dist_ids = np.stack(_batch_dist_ids)
        batch_inputs['distance_ids'] = torch.from_numpy(batch_dist_ids)
        return batch_inputs


    def tp_transform(self, examples):
        batch_text = examples['text']
        batch_ents = examples['ents']
        max_length = self.get_batch_max_length(batch_text=batch_text)
        batch_inputs = self.tokenizer(batch_text,
                                      max_length=max_length,
                                      padding='max_length',
                                      truncation=True,
                                      return_token_type_ids=False,
                                      return_tensors='pt')
        batch_tag_ids = []
        for i, text in enumerate(batch_text):
            ents = batch_ents[i]
            tag_ids = torch.zeros(len(self.ent_labels), max_length, max_length)
            for ent in ents :
                if len(ent['indices']) == 0:
                    log.warn(f'found empty entity indexes in {text}')
                    continue
                start_char = ent['indices'][0]
                start = batch_inputs.char_to_token(i, start_char)
                end_char = ent['indices'][-1] 
                end = batch_inputs.char_to_token(i, end_char)
                label_id = self.ent2id[ent['label']]
                tag_ids[label_id,  start, end] = 1
            batch_tag_ids.append(tag_ids)
        batch_tag_ids = torch.stack(batch_tag_ids, dim=0)
        batch_inputs['tag_ids'] = batch_tag_ids
        return batch_inputs
    
    
    def bio_transform(self, examples):
        batch_text = examples['text']  
        max_length = self.get_batch_max_length(batch_text=batch_text)
        inputs = self.tokenizer(batch_text,
                                padding='max_length',
                                max_length=max_length,
                                truncation=True,
                                return_token_type_ids=False,
                                return_tensors='pt')
        batch_mask = inputs['attention_mask']
        batch_tags = torch.zeros_like(batch_mask, dtype=torch.long)
        batch_tags = batch_tags + (batch_mask - 1) * 100 # 将pad的部分改为-100
        for i, text in enumerate(batch_text):
            ents = examples['ents'][i]
            for ent in ents:
                indices = ent['indices']
                start_char = indices[0]
                start_token = inputs.char_to_token(i, start_char)
                batch_tags[i][start_token] = self.bio2id['B' + '-' + ent['label']]
                if len(indices) > 1:
                    end_char = indices[-1]
                    end_token = inputs.char_to_token(i, end_char)
                    batch_tags[i][start_token: end_token+1] = self.bio2id['I' + '-' + ent['label']]
        inputs['tag_ids'] = batch_tags
        return inputs


import torch
from torch import Tensor
import numpy as np


class MultiLabelCategoricalCrossEntropy(torch.nn.Module):
    '''用于多标签分类的交叉熵
    说明:
    - 阈值应为0
    - y_true和y_pred的shape一致,y_true的元素非0即1,1表示对应的类为目标类,0表示对应的类为非目标类;
    - 请保证y_pred的值域是全体实数;换言之一般情况下y_pred不用加激活函数,尤其是不能加sigmoid或者softmax;
    - 预测阶段则输出y_pred大于0的类;
    参考:
    - https://kexue.fm/archives/7359 
    - https://github.com/bojone/bert4keras/blob/4dcda150b54ded71420c44d25ff282ed30f3ea42/bert4keras/backend.py#L250
    '''
    def __init__(self):
        super().__init__()

    def forward(self, y_pred, y_true):
        """
            y_pred : [batch_size * num_classes, seq_len * seq_len]
            y_true : [batch_size * num_classes, seq_len * seq_len]
        """
        y_pred = (1 - 2 * y_true) * y_pred  # -1 -> pos classes, 1 -> neg classes
        y_pred_neg = y_pred - y_true * 1e7  # mask the pred outputs of pos classes
        y_pred_pos =  y_pred - (1 - y_true) * 1e7  # mask the pred outputs of neg classes
        
        zeros = torch.zeros_like(y_pred[..., :1])
        
        y_pred_neg = torch.cat([y_pred_neg, zeros], dim=-1)
        y_pred_pos = torch.cat([y_pred_pos, zeros], dim=-1)
        neg_loss = torch.logsumexp(y_pred_neg, dim=-1)
        pos_loss = torch.logsumexp(y_pred_pos, dim=-1)
        loss = neg_loss + pos_loss
        if len(loss.shape) == 3:
            loss = loss.sum(dim=1).mean()
        else:
            loss = loss.mean()
        return loss

class SparseMultiLabelCrossEntropy(torch.nn.Module):
    """token-pair稀疏版多标签分类的交叉熵, y_true只传正例
    - y_true.shape=[batch, class, max_instance],
    - y_pred.shape=[batch, class,];
    - 请保证y_pred的值域是全体实数，换言之一般情况下y_pred不用加激活函数，尤其是不能加sigmoid或者softmax；
    - 预测阶段则输出y_pred大于0的类；
    参考:
    - https://kexue.fm/archives/7359 。公式5
    - https://github.com/bojone/bert4keras/blob/4dcda150b54ded71420c44d25ff282ed30f3ea42/bert4keras/backend.py#L272
    """
    def __init__(self, mask_zero: bool = True, eposilon: float = 1e-7, inf=1e7):
        super().__init__()
        self.eposilon = eposilon
        self.mask_zero = mask_zero
        self.inf = inf
        
    def forward(self, y_pred: Tensor, y_true: Tensor):
        zeros = torch.zeros_like(y_pred[..., :1])
        y_pred = torch.cat([y_pred, zeros], dim=-1) # 最后添加一个0 
        if self.mask_zero:
            infs = zeros + self.inf
            y_pred = torch.cat([infs, y_pred[..., 1:]], dim=-1)
        y_pos_2 = torch.gather(y_pred, index=y_true, dim=-1)
        y_pos_1 = torch.cat([y_pos_2, zeros], dim=-1)
        if self.mask_zero:
            y_pred = torch.cat([-infs, y_pred[..., 1:]], dim=-1)
            y_pos_2 = torch.gather(y_pred, index=y_true, dim=-1)
            
        pos_loss = torch.logsumexp(-y_pos_1, dim=-1)
        # 计算负类对应损失
        all_loss = torch.logsumexp(y_pred, dim=-1) # 公式里面的a
        aux_loss = torch.logsumexp(y_pos_2, dim=-1) - all_loss # b - a
        aux_loss = torch.clamp(1 - torch.exp(aux_loss), min=self.eposilon, max=1) # 1-exp(b-a)
        neg_loss = all_loss + torch.log(aux_loss) # a + log[1-exp(b-a)]
        loss = pos_loss + neg_loss
        return loss.sum(dim=1).mean()
    
        # pos_loss = torch.logsumexp(-1 * torch.cat([torch.gather(y_pred, index=y_true, dim=-1), torch.zeros_like(y_pred[..., :1])], dim=-1), dim=-1)
        # a = torch.logsumexp(torch.cat([y_pred, torch.zeros_like(y_pred[..., :0])], dim=-1), dim=-1)
        # b = torch.logsumexp(torch.gather(y_pred, index=y_true, dim=-1), dim=-1)
        # neg_loss = a + torch.log(torch.clamp(1 - torch.exp(b - a), min=self.eposilon, max=1))
        # loss = (pos_loss + neg_loss).sum(dim=1).mean()
        # return loss
    
    
class CoSentLoss(torch.nn.Module):
    '''CoSentLoss的实现
    参考:
    - https://kexue.fm/archives/8847
    '''
    def __init__(self) -> None:
        super().__init__()
    
    def forward(self, preds, targs, device):
        '''
        参数:
        - preds: 余弦相似度 [0.1, 0.2, 0.3,...]
        - targs: 标签 [0,1,1,0, ...]
        - device: 运行的设备
        返回:
        - loss: 余弦相似度与标签的交叉熵
        '''
        # 20为公式里面的lambda 这里取20
        preds = preds * 20
        # 利用广播机制, 所有位置, 两两差值
        preds = preds[:, None] - preds[None, :]
        # 
        targs = targs[:, None] < targs[None, :]
        targs = targs.float()
        # 这里之所以要这么减，是因为公式中所有的正样本对的余弦值减去负样本对的余弦值才计算损失，把不是这些部分通过exp(-inf)忽略掉
        preds = preds - (1 - targs) * 1e12
        preds = preds.view(-1)
        preds = torch.cat((torch.tensor([0], dtype=torch.float, device=device), preds), dim=0)
        return torch.logsumexp(preds, dim=0)


from functools import lru_cache
import torch
from ..utils.make_datamodule import PLMBaseDataModule


class TokenClassificationDataModule(PLMBaseDataModule):
    '''序列标注数据模块
    数据集字段:
    - tokens: list[str], 例如: ['我', '是', '一', '中', '国', '人', '。']
    - labels: list[str], 例如: ['O', 'O', 'O', 'B-LOC', 'I-LOC', 'O', 'O']
    '''
    def __init__(
        self,
        dataset: str,
        plm: str,
        max_length: int,
        batch_size: int,
        pin_memory: bool,
        num_workers: int,
        pretrained_dir: str ,
        data_dir: str ,
        label_pad_id: int = -100
        ):
        super().__init__()

        self.save_hyperparameters()


    def transform(self, examples):
        batch = {'inputs':[], 'label_ids':[]}
        for i, t_list in enumerate(examples['tokens']):
            tokens = [t['text'] for t in t_list]
            label_ids = [self.hparams['label2id'][t['label']] for t in t_list]
            inputs = self.tokenizer.encode_plus(
                tokens,  
                add_special_tokens=True,
                padding='max_length',  
                max_length=self.hparams.max_length,
                truncation=True)
            inputs = dict(zip(inputs.keys(), map(torch.tensor, inputs.values())))
            if len(label_ids) < (self.hparams.max_length-2):
                O_id = self.hparams.label2id['O']
                label_ids = [O_id] + label_ids + [O_id] + [-100]*(self.hparams.max_length-len(label_ids)-2)
            else: label_ids = [O_id] + label_ids + [O_id]
            batch['inputs'].append(inputs)
            label_ids = torch.tensor(label_ids)
            batch['label_ids'].append(label_ids)
        return batch

    
    @property
    @lru_cache()
    def label2id(self):
        set_labels = sorted(set([t['label'] for t_list in self.dataset['train']['tokens'] for t in t_list]))
        label2id = {label: i for i, label in enumerate(set_labels)}
        return label2id


    def setup(self, stage):
        self.hparams.label2id = self.label2id
        self.hparams.id2label = {i:l for l,i in self.label2id.items()}
        self.dataset.set_transform(transform=self.transform)



import math
import torch
import torch.nn.functional as F

class GELU(torch.nn.Module):
    """
    高斯误差线性单元,该激活函数在激活中加入了**随机正则**的思想
    参考:
    - https://kexue.fm/archives/7309
    - https://arxiv.org/abs/1606.08415
    """
    def forward(self, x):
        return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))


class  GELU_Approximate(torch.nn.Module):
    """ 
    gelu激活函数的近似版本
    """
    def forward(self, x):
        return 0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))


class SWISH(torch.nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(x)



activations = {'gelu':GELU, 'swish':SWISH,'gelu_approximate':GELU_Approximate}




from typing import List
from ...utils.make_model import PLMBaseModel
from ...layers.dropout import MultiDropout
from ...metrics.triple import TripleF1, Triple
import torch.nn as nn
import torch


class CasRelLoss(nn.BCELoss):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        

    def forward(self, subs_x, subs_y, objs_x, objs_y, mask):

        # sujuect部分loss
        subject_loss = super().forward(subs_x, subs_y)
        subject_loss = subject_loss.mean(dim=-1)
        subject_loss = (subject_loss * mask).sum() / mask.sum()
        # object部分loss
        object_loss = super().forward(objs_x, objs_y)
        object_loss = object_loss.mean(dim=-1).sum(dim=-1)
        object_loss = (object_loss * mask).sum() / mask.sum()
        return subject_loss + object_loss



class CasRelForRelationExtraction(PLMBaseModel):
    """CasRel 关系抽取模型, 无法解决嵌套实体的情况

    Args:
        - lr: 学习率
        - dropout: dropout的比率
        - scheduler: 学习率衰减策略必须
        - threshold: 阈值
        
    Reference:
        [1] https://github.com/xiangking/ark-nlp/blob/main/ark_nlp/model/re/prgc_bert/prgc_bert.py
    """
    
    def __init__(self,
                 lr: float,
                 weight_decay: float = 0.01,
                 scheduler: str = 'linear_warmup_step',
                 threshold: float = 0.5,
                 **kwargs):
        super().__init__()

        self.plm = self.get_plm_architecture()
        
        self.dropout = MultiDropout()
        # self.condLayerNorm = LayerNorm(hidden_size=self.plm.config.hidden_size, conditional_size=self.plm.config.hidden_size*2)
        self.sub_classifier = nn.Linear(self.plm.config.hidden_size, 2)
        self.obj_classifier = nn.Linear(self.plm.config.hidden_size, len(self.hparams.label2id)*2)

        self.sub_criterion = nn.BCEWithLogitsLoss()
        self.obj_criterion = nn.BCEWithLogitsLoss()
        
        self.triple_metric = TripleF1()
        self.sub_metric = SpanOffsetF1()
        

    def forward(self,
                input_ids,
                attention_mask=None,
                token_type_ids=None,
                **kwargs):
        batch_size, seq_len = input_ids.shape
        batch_lengths = torch.sum(attention_mask, -1)
        batch_hidden_state = self.encode_step(input_ids, token_type_ids, attention_mask)
        batch_sub_logits = self.sub_classifier_step(batch_hidden_state)
        batch_sub_scores = torch.sigmoid(batch_sub_logits)
        batch_sub_spans = self.get_sub_spans(batch_sub_scores, attention_mask, self.hparams.threshold)
        batch_relations = []
        for i in range(batch_size):
            sub_spans = batch_sub_spans[i]
            length = batch_lengths[i]
            hidden_state = batch_hidden_state[i].unsqueeze(0).repeat(len(sub_spans), 1, 1) # 改变回原来形状batch, seq, hidden
            if sub_spans:
                # sub_spans : num_spans, 2
                sub_spans = torch.tensor(sub_spans, dtype=torch.long, device=self.device)
                obj_logits = self.obj_classifier_step(hidden_state, sub_spans)
                relations = self.get_triples(obj_logits, sub_spans, length)
                batch_relations.append(relations)
            else:
                batch_relations.append(set())
        return batch_relations, batch_sub_spans



        
    def training_step(self, batch, batch_idx):
        input_ids = batch['input_ids']
        token_type_ids = batch['token_type_ids']
        attention_mask = batch['attention_mask']
        subs = batch['subs']
        sub = batch['sub']
        objs = batch['objs']
        last_hidden_state = self.encode_step(input_ids, token_type_ids, attention_mask)
        sub_logits = self.sub_classifier_step(last_hidden_state)
        obj_logits = self.obj_classifier_step(last_hidden_state, sub)
        subject_loss = self.sub_criterion(sub_logits, subs)
        subject_loss = subject_loss.mean(dim=-1)
        subject_loss = (subject_loss * attention_mask).sum() / attention_mask.sum()
        object_loss = self.obj_criterion(obj_logits, objs)
        object_loss = object_loss.mean(dim=-1).sum(dim=-1)
        object_loss = (object_loss * attention_mask).sum() / attention_mask.sum()
        loss = subject_loss + object_loss
        self.log('train/loss', loss)
        return {'loss':loss}

        
    def validation_step(self, batch, batch_idx):
        input_ids = batch['input_ids']
        batch_size, seq_len = input_ids.shape
        token_type_ids = batch['token_type_ids']
        attention_mask = batch['attention_mask']
        batch_objs = batch['objs']
        batch_sub = batch['sub']
        batch_subs_true = self.get_sub_spans(batch['subs'], attention_mask, self.hparams.threshold)
        batch_relations, batch_subs = self(input_ids, token_type_ids, attention_mask)
        batch_triples = []
        for i in range(len(batch_relations)):
            triples = batch_relations[i]
            ts = set()
            for t in triples:
                ts.add(Triple(t))
            batch_triples.append(ts)
        batch_triples_true = []
        for i in range(batch_size):
            triples = self.get_triples(batch_objs[i].unsqueeze(0), batch_sub[i].unsqueeze(0), torch.sum(attention_mask[i]))
            batch_triples_true.append(set([Triple(t) for t in triples]))
                        
        self.triple_metric(batch_triples, batch_triples_true)
        self.sub_metric(batch_subs, batch_subs_true)
        self.log('triple/f1', self.triple_metric, on_step=False, on_epoch=True, prog_bar=True)
        self.log('sub/f1', self.sub_metric, on_step=False, on_epoch=True, prog_bar=True)


        
    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.plm.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.plm.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.sub_classifier.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.sub_classifier.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.obj_classifier.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.obj_classifier.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr, 'weight_decay': 0.0}
        ]
        optimizer = torch.optim.AdamW(grouped_parameters, eps=1e-5)
        scheduler_config = self.get_scheduler_config(optimizer=optimizer, name=self.hparams.scheduler)
        return [optimizer], [scheduler_config]
    
    def encode_step(self, input_ids, token_type_ids, attention_mask):
        last_hidden_state = self.plm(input_ids, token_type_ids, attention_mask).last_hidden_state
        return last_hidden_state
    
    
    def sub_classifier_step(self, last_hidden_state):
        # sub_logits = (torch.sigmoid(self.sub_classifier(last_hidden_state)))**2
        sub_logits = self.sub_classifier(last_hidden_state)
        # sub_scores = torch.sigmoid(sub_logits)
        return sub_logits

    
    def obj_classifier_step(self, last_hidden_state, sub_spans):
        sub_reps = self.get_sub_representation(last_hidden_state, sub_spans)
        # output = self.condLayerNorm([last_hidden_state, sub_reps])
        # output = (torch.sigmoid(self.obj_classifier(output)))**4
        output = last_hidden_state + sub_reps
        output =self.obj_classifier(output)
        obj_logits = output.reshape(*output.shape[:2], len(self.hparams.label2id), 2)
        return obj_logits
    
    
    @staticmethod
    def get_sub_representation(hidden_states, sub_spans):
        """根据sub从output中取出subject的向量表征
        hidden_state: batch, seq, hidden
        sub_spans: batch, num_subs, 2

        """
        start = torch.gather(hidden_states, dim=1, index=sub_spans[:, :1].unsqueeze(2).expand(-1, -1, hidden_states.shape[-1]))
        end = torch.gather(hidden_states, dim=1, index=sub_spans[:, 1:].unsqueeze(2).expand(-1, -1, hidden_states.shape[-1]))
        # subject = torch.cat([start, end], 2)
        sub_rep = (start + end) / 2
        
        return sub_rep
        

    def get_sub_spans(self, sub_scores, mask, threshold) -> List:
        """将sub classifier 线性层经过sigmoid输出的score转换为sub spans
        返回一个列表
        """
        lengths = torch.sum(mask, -1)
        starts = sub_scores[:, :, :1]
        ends = sub_scores[:, :, 1:]
        sub_spans = []
        for start, end, l in zip(starts, ends, lengths):
            tmp = set()
            start = start.squeeze()[:l]
            end = end.squeeze()[:l]
            for i, st in enumerate(start):
                if st > threshold:
                    s = i
                    for j in range(i, l):
                        if end[j] > threshold:
                            e = j
                            if (s,e) not in sub_spans:
                                tmp.add((s,e))
                            break

            sub_spans.append(tmp)
        return sub_spans

    def get_triples(self, obj_logits, sub_spans, length):
        """_summary_

        Args:
            obj_logits (_type_): num_subs, seq_len, num_rels, 2
            sub_spans (_type_): num_subs, 2
            length (_type_): _description_

        Returns:
            _type_: _description_
        """
        num_label = obj_logits.shape[2]
        num_subs = len(sub_spans)
        triples = set()
        subjects = set()
        # objects = []
        for b in range(num_subs):
            tmp = obj_logits[b, ...]
            start, end = sub_spans[b]
            subject = (start.item(), end.item())
            if subject not in subjects:
                subjects.add(subject)
                for label_id in range(num_label):
                    start = tmp[:, label_id, :1]
                    end = tmp[:, label_id, 1:]
                    start = start.squeeze()[:length]
                    end = end.squeeze()[:length]
                    for i, st in enumerate(start):
                        if st > self.hparams.threshold:
                            s = i
                            for j in range(i, length):
                                if end[j] > 0.5:
                                    e = j
                                    object = (s,e)
                                    # if object not in objects:
                                    #     objects.append(object)
                                    if (*subject, label_id, *object) not in triples:
                                        triples.add((*subject, label_id, *object))
                                    break
        # print(triples) 
        return triples


from pydantic import BaseModel
from typing import Tuple, Optional
from torchmetrics import Metric
import torch
from typing import List, Tuple, Set
from ..data.doc import Entity, Relation

    
class RelationF1(Metric):
    """实体和关系联合抽取的F1指标
    """

    full_state_update: Optional[bool] = False

    def __init__(self):
        super().__init__()
        self.add_state('correct', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_pred', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_true', default=torch.tensor(0.0), dist_reduce_fx='sum')

    def update(self, pred: List[Set[Relation]], true: List[Set[Relation]]) -> None:
        """
        参数:
        - pred: batch_size大小的列表, 列表内为每个样本的预测的三元组的集合(set)
        - true: 同pred
        """
        assert len(pred) == len(true), f'pred : {pred}, true : {true}'
        for p, t in zip(pred, true):
            self.correct += len(p & t)
            self.all_pred += len(p)
            self.all_true += len(t)
    
    def compute(self):
        # 下面这个公式可以通过化简得到
        return 2 * self.correct / (self.all_pred + self.all_true + 1e-10)


from functools import lru_cache
from ..utils.make_datamodule import char_idx_to_token, get_logger, PLMBaseDataModule, sequence_padding
import torch
from typing import Dict, Union, List
import numpy as np
import pandas as pd
import random


log = get_logger()


class RelationExtractionDataModule(PLMBaseDataModule):
    """三元组抽取数据模块,
        数据集样式:
            {'text':'小明是小花的朋友'。
            'rels': [{'o': {'indices': [0, 1, 2], 'text': '小明', 'label': '人物'},'p': '朋友','s': {'indices': [3, 4, 5], 'text': '小花', 'label': '人物'}}]}}
        数据集说明:
            当只抽取三元组的时候subject,object中的标签可以为''
        """
    def __init__(self, 
                 dataset: str, 
                 batch_size: int ,
                 plm: str = 'hfl/chinese-roberta-wwm-ext',
                 **kwargs):
        """
        Args:
            dataset (str): 数据集名称
            plm (str): 预训练模型名称
            auto_length (str, int): 自动设置最大长度的策略, 可以为'max', 'mean'或者>0的数,超过512的设置为512
            batch_size (int): 训练,验证,测试数据集的批次大小,
            num_workers (int): 多进程数
            pin_memory (bool): 是否应用锁页内存,
            dataset_dir (str): 数据集默认路径
            plm_dir (str): 预训练模型默认路径
        """
        super().__init__()


    def setup(self, stage: str = 'fit') -> None:
        """对数据设置转换"""
        self.hparams.id2rel = self.id2rel
        self.hparams.id2combined = self.id2combined
        self.hparams.id2onerel = self.id2onerel
        self.hparams.id2ent = self.id2ent
        
    @property
    @lru_cache()
    def combined_labels(self) -> List:
        """将主体客体与其实体标签结合起来, (主体, 地点)
        """
        def get_labels(e):
            return ['主体'+'-'+e['s']['label'], '客体'+'-'+ e['o']['label']]
        labels = set(np.concatenate(pd.Series(np.concatenate(self.train_df.rels)).apply(get_labels)))
        ls = [tuple(l.split('-')) for l in labels]
        return ls
    
    @property
    @lru_cache()
    def combined2id(self) -> Dict:
        return {l:i for i,l in enumerate(self.combined_labels)}
    
    @property
    def id2combined(self) -> Dict:
        return {i:l for l,i in self.combined2id.items()}
    
    @property
    @lru_cache()
    def ent_labels(self) -> List:
        def get_labels(e):
            return [e['s']['label'], e['o']['label']]
        return list(set(np.concatenate(pd.Series(np.concatenate(self.train_df.rels)).apply(get_labels))))
    
    @property
    def ent2id(self) -> Dict:
        return {l:i for i,l in enumerate(self.ent_labels)}
    
    @property
    def id2ent(self) -> Dict:
        return {i:l for l,i in self.ent2id.items()}
    
    
    @property
    @lru_cache()
    def rel_labels(self) -> List:
        labels = set(pd.Series(np.concatenate(self.train_df.rels.values)).apply(lambda x: x['p']))
        return list(labels)

    @property
    def rel2id(self) -> Dict:
        label2id = {label: i for i, label in enumerate(self.rel_labels)}
        return label2id
    
    @property
    def id2rel(self) -> Dict:
        return {i:l for l,i in self.rel2id.items()}
    
    @property
    def onerel2id(self):
        return {'O':0, 'HB-TB':1, 'HB-TE':2, 'HE-TE':3}
    
    @property
    def id2onerel(self) -> Dict:
        return {i:l for l,i in self.onerel2id.items()}
    
    def sparse_triple_transform(self, examples):
        batch_text = examples['text']
        max_length = self.get_batch_max_length(batch_text=batch_text)
        batch_inputs = self.tokenizer(batch_text,
                                      padding=True,
                                      truncation=True,
                                      max_length=max_length,
                                      return_tensors='pt',
                                      return_token_type_ids=False)
        batch_so_tags = []
        batch_head_tags = []
        batch_tail_tags = []
        batch_triples = examples['rels']
        for i, text in enumerate(batch_text):
            triples = batch_triples[i]
            so_tags = [set() for _ in range(2)]
            head_tags = [set() for _ in range(len(self.rel_labels))]
            tail_tags = [set() for _ in range(len(self.rel_labels))]
            for triple in triples:
                try:
                    sub_head = triple['s']['indices'][0]
                    _sub_head = batch_inputs.char_to_token(i, sub_head)
                    sub_tail = triple['s']['indices'][-1]
                    _sub_tail = batch_inputs.char_to_token(i, sub_tail)
                    sub_label = triple['s']['label']
                    assert _sub_head is not None and _sub_tail is not None
                except:
                    log.warning(f'subject {(sub_head, sub_tail)} align fail in \n {text}')
                    continue
                
                try:
                    obj_head = triple['o']['indices'][0]
                    obj_tail = triple['o']['indices'][-1]
                    obj_label = triple['o']['label']
                    _obj_head = batch_inputs.char_to_token(i, obj_head)
                    _obj_tail = batch_inputs.char_to_token(i, obj_tail)
                    assert _obj_head is not None and _obj_tail is not None
                except:
                    log.warning(f'object {(obj_head, obj_tail)} align fail in \n {text}')
                    continue
                
                so_tags[0].add((_sub_head, _sub_tail))
                so_tags[1].add((_obj_head, _obj_tail))
                head_tags[self.rel2id[triple['p']]].add((_sub_head, _obj_head))
                tail_tags[self.rel2id[triple['p']]].add((_sub_tail, _obj_tail))
            for tag in so_tags + head_tags + tail_tags:
                if not tag:
                    tag.add((0,0))
            so_tags = sequence_padding([list(l) for l in so_tags])
            head_tags = sequence_padding([list(l) for l in head_tags])
            tail_tags = sequence_padding([list(l) for l in tail_tags])
            batch_so_tags.append(so_tags)
            batch_head_tags.append(head_tags)
            batch_tail_tags.append(tail_tags)
        batch_inputs['so_tags'] = torch.tensor(sequence_padding(batch_so_tags, seq_dims=2), dtype=torch.long)
        batch_inputs['head_tags'] = torch.tensor(sequence_padding(batch_head_tags, seq_dims=2), dtype=torch.long)
        batch_inputs['tail_tags'] = torch.tensor(sequence_padding(batch_tail_tags, seq_dims=2), dtype=torch.long)
        return batch_inputs
    
        
    def triple_transform(self, example) -> Dict:
        batch_text = example['text']
        batch_triples = example['rels']
        batch_so_ids = []
        batch_head_ids = []
        batch_tail_ids = []
        max_length = self.get_batch_max_length(batch_text)
        batch_inputs = self.tokenizer(batch_text, 
                                      max_length=max_length,
                                      padding=True,
                                      truncation=True,
                                      return_token_type_ids=False,
                                      return_tensors='pt')
        for i, text in enumerate(batch_text):
            triples = batch_triples[i]
            so_ids = torch.zeros(2, max_length, max_length, dtype=torch.long)
            head_ids = torch.zeros(len(self.rel_labels), max_length, max_length, dtype=torch.long)
            tail_ids = torch.zeros(len(self.rel_labels), max_length, max_length, dtype=torch.long)
            for triple in triples:
                try:
                    sub_start = triple['s']['indices'][0]
                    sub_end = triple['s']['indices'][-1]
                    _sub_start = batch_inputs.char_to_token(i, sub_start)
                    _sub_end = batch_inputs.char_to_token(i, sub_end)
                    obj_start = triple['o']['indices'][0]
                    obj_end = triple['o']['indices'][-1]
                    _obj_start = batch_inputs.char_to_token(i, obj_start)
                    _obj_end = batch_inputs.char_to_token(i, obj_end)
                    so_ids[0][_sub_start][_sub_end] = 1
                    so_ids[1][_obj_start][_obj_end] = 1
                    head_ids[self.rel2id[triple['p']]][_sub_start][_obj_start] = 1
                    tail_ids[self.rel2id[triple['p']]][_sub_end][_obj_end] = 1
                except:
                    log.warning(f'sub char offset {(sub_start, sub_end)} or obj char offset {(obj_start, obj_end)} align to token offset failed in \n{text}')
                    pass
            batch_so_ids.append(so_ids)
            batch_head_ids.append(head_ids)
            batch_tail_ids.append(tail_ids)
        batch_inputs['so_tags'] = torch.stack(batch_so_ids, dim=0)
        batch_inputs['head_tags'] = torch.stack(batch_head_ids, dim=0)
        batch_inputs['tail_tags'] = torch.stack(batch_tail_ids, dim=0)
        return batch_inputs
    
    
    def onerel_transform(self, example):
        texts = example['text']
        batch_triples = example['rels']
        batch_inputs = {'input_ids': [], 'attention_mask': [], 'token_type_ids': []}
        batch_tag_ids = []
        batch_loss_mask = []
        max_length = self.hparams.max_length
        for i, text in enumerate(texts):
            inputs = self.tokenizer(text, 
                                    padding='max_length',  
                                    max_length=max_length,
                                    truncation=True,
                                    return_offsets_mapping=True)
            batch_inputs['input_ids'].append(inputs['input_ids'])
            batch_inputs['attention_mask'].append(inputs['attention_mask'])
            batch_inputs['token_type_ids'].append(inputs['token_type_ids'])
            offset_mapping = inputs['offset_mapping']
            triples = batch_triples[i]
            tag_ids = torch.zeros(len(self.label2id), max_length, max_length, dtype=torch.long)
            loss_mask = torch.ones(1, max_length, max_length, dtype=torch.long)
            att_mask = torch.tensor(inputs['attention_mask'])
            loss_mask = loss_mask * att_mask.unsqueeze(0) * att_mask.unsqueeze(0).T
            for triple in triples:
                try:
                    rel = triple['p']
                    rel_id = self.label2id[rel]
                    sub_start = triple['s']['indices'][0]
                    sub_end = triple['s']['indices'][-1]
                    sub_start = char_idx_to_token(sub_start, offset_mapping=offset_mapping)
                    sub_end = char_idx_to_token(sub_end, offset_mapping=offset_mapping)
                    obj_start = triple['o']['indices'][0]
                    obj_end = triple['o']['indices'][-1]
                    obj_start = char_idx_to_token(obj_start, offset_mapping=offset_mapping)
                    obj_end = char_idx_to_token(obj_end, offset_mapping=offset_mapping)
                    if sub_start != sub_end and obj_start != obj_end:
                        tag_ids[rel_id][sub_start][obj_start] = self.onerel_tag2id['HB-TB']
                        tag_ids[rel_id][sub_start][obj_end] = self.onerel_tag2id['HB-TE']
                        tag_ids[rel_id][sub_end][obj_end] = self.onerel_tag2id['HE-TE']
                    if sub_start == sub_end and obj_start != obj_end:
                        tag_ids[rel_id][sub_start][obj_start] = self.onerel_tag2id['HB-TB']
                        tag_ids[rel_id][sub_end][obj_end] = self.onerel_tag2id['HE-TE']
                    if sub_start != sub_end and obj_start == obj_end:
                        tag_ids[rel_id][sub_start][obj_start] = self.onerel_tag2id['HB-TB']
                        tag_ids[rel_id][sub_end][obj_end] = self.onerel_tag2id['HE-TE']
                    if sub_start == sub_end and obj_start == obj_end:
                        tag_ids[rel_id][sub_start][obj_start] = self.onerel_tag2id['HB-TB']
                except Exception as e:
                    log.exception(e)
                    log.warning(f'sub char offset {(sub_start, sub_end)} or obj char offset {(obj_start, obj_end)} align to token offset failed in \n{text}')
                    pass
            batch_tag_ids.append(tag_ids)
            batch_loss_mask.append(loss_mask)
        batch = dict(zip(batch_inputs.keys(), map(torch.tensor, batch_inputs.values())))
        batch['tag_ids'] = torch.stack(batch_tag_ids, dim=0)
        batch['loss_mask'] = torch.stack(batch_loss_mask, dim=0)
        return batch
    
    
    def casrel_transform(self, example):
        texts = example['text']
        batch_triples = example['rels']
        batch_inputs = {'input_ids': [], 
                        'attention_mask': [], 
                        'token_type_ids': []}
        batch_subs = []
        batch_objs = []
        batch_sub = []
                        
        max_length = self.hparams.max_length
        for i, text in enumerate(texts):
            inputs = self.tokenizer(text, 
                                    padding='max_length',  
                                    max_length=max_length,
                                    truncation=True,
                                    return_offsets_mapping=True)
            offset_mapping = inputs['offset_mapping']
            triples = batch_triples[i]
            s2ro_map = {}
            for triple in triples:
                sub_start = triple['s']['indices'][0]
                sub_end = triple['s']['indices'][-1]
                sub_start = char_idx_to_token(sub_start, offset_mapping=offset_mapping)
                sub_end = char_idx_to_token(sub_end, offset_mapping=offset_mapping)
                obj_start = triple['o']['indices'][0]
                obj_end = triple['o']['indices'][-1]
                obj_start = char_idx_to_token(obj_start, offset_mapping=offset_mapping)
                obj_end = char_idx_to_token(obj_end, offset_mapping=offset_mapping)
                rel_id = self.hparams.label2id[triple['p']]
                if (sub_start, sub_end) not in s2ro_map:
                    s2ro_map[(sub_start, sub_end)] = [] 
                s2ro_map[(sub_start, sub_end)].append((obj_start, obj_end, rel_id))
                
            if s2ro_map:
                subs = torch.zeros(max_length, 2, dtype=torch.float)
                for s in s2ro_map:
                    subs[s[0], 0] = 1
                    subs[s[1], 1] = 1
                # for sub_head_idx, sub_tail_idx in list(s2ro_map.keys()):
                sub_head_idx, sub_tail_idx = random.choice(list(s2ro_map.keys()))
                sub = torch.tensor([sub_head_idx, sub_tail_idx], dtype=torch.long)
                objs = torch.zeros((max_length, len(self.hparams.label2id), 2), dtype=torch.float)
                for ro in s2ro_map.get((sub_head_idx, sub_tail_idx), []):
                    objs[ro[0], ro[2], 0] = 1
                    objs[ro[1], ro[2], 1] = 1
                batch_subs.append(subs)
                batch_objs.append(objs)
                batch_sub.append(sub)
                batch_inputs['input_ids'].append(inputs['input_ids'])
                batch_inputs['attention_mask'].append(inputs['attention_mask'])
                batch_inputs['token_type_ids'].append(inputs['token_type_ids'])
        batch = dict(zip(batch_inputs.keys(), map(torch.tensor, batch_inputs.values())))
        batch['subs'] = torch.stack(batch_subs, dim=0)
        batch['sub'] = torch.stack(batch_sub, dim=0)
        batch['objs'] = torch.stack(batch_objs, dim=0)
        return batch
        

    def sparse_combined_transform(self, examples):
        batch_text = examples['text']
        max_length = self.get_batch_max_length(batch_text=batch_text)
        batch_inputs = self.tokenizer(batch_text,
                                      padding=True,
                                      truncation=True,
                                      max_length=max_length,
                                      return_tensors='pt',
                                      return_token_type_ids=False)
        batch_combined_tags = []
        batch_head_tags = []
        batch_tail_tags = []
        batch_triples = examples['rels']
        for i, text in enumerate(batch_text):
            triples = batch_triples[i]
            combined_tags = [set() for _ in range(len(self.combined_labels))]
            head_tags = [set() for _ in range(len(self.rel_labels))]
            tail_tags = [set() for _ in range(len(self.rel_labels))]
            for triple in triples:
                try:
                    sub_head = triple['s']['indices'][0]
                    _sub_head = batch_inputs.char_to_token(i, sub_head)
                    sub_tail = triple['s']['indices'][-1]
                    _sub_tail = batch_inputs.char_to_token(i, sub_tail)
                    sub_label = triple['s']['label']
                    assert _sub_head is not None and _sub_tail is not None
                except:
                    log.warning(f'subject {(sub_head, sub_tail)} align fail in \n {text}')
                    continue
                
                try:
                    obj_head = triple['o']['indices'][0]
                    obj_tail = triple['o']['indices'][-1]
                    obj_label = triple['o']['label']
                    _obj_head = batch_inputs.char_to_token(i, obj_head)
                    _obj_tail = batch_inputs.char_to_token(i, obj_tail)
                    assert _obj_head is not None and _obj_tail is not None
                except:
                    log.warning(f'object {(obj_head, obj_tail)} align fail in \n {text}')
                    continue
                
                combined_tags[self.combined2id[('主体', sub_label)]].add((_sub_head, _sub_tail))
                combined_tags[self.combined2id[('客体', obj_label)]].add((_obj_head, _obj_tail))
                head_tags[self.rel2id[triple['p']]].add((_sub_head, _obj_head))
                tail_tags[self.rel2id[triple['p']]].add((_sub_tail, _obj_tail))
            for tag in combined_tags + head_tags + tail_tags:
                if not tag:
                    tag.add((0,0))
            combined_tags = sequence_padding([list(l) for l in combined_tags])
            head_tags = sequence_padding([list(l) for l in head_tags])
            tail_tags = sequence_padding([list(l) for l in tail_tags])
            batch_combined_tags.append(combined_tags)
            batch_head_tags.append(head_tags)
            batch_tail_tags.append(tail_tags)
        batch_inputs['combined_tags'] = torch.tensor(sequence_padding(batch_combined_tags, seq_dims=2), dtype=torch.long)
        batch_inputs['head_tags'] = torch.tensor(sequence_padding(batch_head_tags, seq_dims=2), dtype=torch.long)
        batch_inputs['tail_tags'] = torch.tensor(sequence_padding(batch_tail_tags, seq_dims=2), dtype=torch.long)
        return batch_inputs
    
    
    def combined_transform(self, examples):
        batch_text = examples['text']
        max_length = self.get_batch_max_length(batch_text=batch_text)
        batch_inputs = self.tokenizer(batch_text,
                                      padding=True,
                                      truncation=True,
                                      max_length=max_length,
                                      return_tensors='pt',
                                      return_token_type_ids=False)
        batch_combined_tags = []
        batch_head_tags = []
        batch_tail_tags = []
        batch_triples = examples['rels']
        for i, text in enumerate(batch_text):
            triples = batch_triples[i]
            combined_tags = torch.zeros(len(self.combined_labels), max_length, max_length, dtype=torch.long)
            head_tags = torch.zeros(len(self.rel_labels), max_length, max_length, dtype=torch.long)
            tail_tags = torch.zeros(len(self.rel_labels), max_length, max_length, dtype=torch.long)
            for triple in triples:
                sub_head = triple['s']['indices'][0]
                _sub_head = batch_inputs.char_to_token(i, sub_head)
                sub_tail = triple['s']['indices'][-1]
                _sub_tail = batch_inputs.char_to_token(i, sub_tail)
                sub_label = triple['s']['label']
                
                obj_head = triple['o']['indices'][0]
                obj_tail = triple['o']['indices'][-1]
                obj_label = triple['o']['label']
                _obj_head = batch_inputs.char_to_token(i, obj_head)
                _obj_tail = batch_inputs.char_to_token(i, obj_tail)
                
                if _sub_head is None or _sub_tail is None :
                    log.warning(f'subject {(sub_head, sub_tail)} align fail in \n {text}')
                    continue
                if _obj_head is None or _obj_tail is None:
                    log.warning(f'object {(obj_head, obj_tail)} align fail in \n {text}')
                    continue
                
                combined_tags[self.combined2id[('主体', sub_label)], _sub_head, _sub_tail] = 1
                combined_tags[self.combined2id[('客体', obj_label)], _obj_head, _obj_tail] = 1
                head_tags[self.rel2id[triple['p']], _sub_head, _obj_head] = 1
                tail_tags[self.rel2id[triple['p']], _sub_tail, _obj_tail] = 1
            batch_combined_tags.append(combined_tags)
            batch_head_tags.append(head_tags)
            batch_tail_tags.append(tail_tags)
        batch_inputs['combined_tags'] = torch.stack(batch_combined_tags, dim=0)
        batch_inputs['head_tags'] = torch.stack(batch_head_tags, dim=0)
        batch_inputs['tail_tags'] = torch.stack(batch_tail_tags, dim=0)
        return batch_inputs


from datasets import Dataset as Ds
from datasets import DatasetDict as DsD
from datasets import load_from_disk
from typing import Union, Tuple


class Dataset(Ds):
    def train_val_split(self, 
                        val_frac: float =0.1,
                        return_dataset_dict: bool =True) -> Union[Tuple["Dataset", "Dataset"], "DatasetDict"]:
        """split dataset into tarin and validation datasets
        Args:
            dataset (Dataset): dataset to split
            val_frac (float, optional): validation radio of all dataset. Defaults to 0.1.
            return_dataset_dict (bool, optional): if return_dataset_dict is True, return a DatasetDict,
                otherwise return a tuple of train, val datasets. Defaults to True.

        Returns:
            Union[Tuple[Dataset, Dataset], DatasetDict]: if return_dataset_dict is True, return a DatasetDict, otherwise return a tuple of train, val datasets
        """
        df = self.to_pandas()
        train_df = df.sample(frac=1-val_frac)
        val_df = df.drop(train_df.index)
        train_ds = self.from_pandas(train_df, preserve_index=False)
        val_ds = self.from_pandas(val_df, preserve_index=False)
        if not return_dataset_dict:
            return train_ds, val_ds
        else:
            return DatasetDict({'train': train_ds, 'validation': val_ds})
        
    
    def train_val_test_split(self,
                             val_frac: float =0.1,
                             test_frac: float =0.1,
                             return_dataset_dict: bool =True) -> Union[Tuple["Dataset", "Dataset", "Dataset"], "DatasetDict"]:
        """split dataset into tarin vlidation and test datasets

        Args:
            dataset (Dataset): dataset to split
            val_frac (float, optional): validation radio of all dataset. Defaults to 0.1.
            test_frac (float, optional): test radio of all dataset. Defaults to 0.1.

        Returns:
            Union[Tuple[Dataset, Dataset, Dataset], DatasetDict]: if return_dataset_dict is True, return a DatasetDict, 
                otherwise return a tuple of train, val, test datasets
            
        """
        df = self.to_pandas()
        train_df = df.sample(frac=1-val_frac-test_frac)
        other_df = df.drop(train_df.index)
        val_df = other_df.sample(frac=1-(test_frac/(val_frac+test_frac)))
        test_df = other_df.drop(val_df.index)
        train_ds = self.from_pandas(train_df, preserve_index=False)
        val_ds = self.from_pandas(val_df, preserve_index=False)
        test_ds = self.from_pandas(test_df, preserve_index=False)
        if not return_dataset_dict:
            return train_ds, val_ds, test_ds
        else:
            return DatasetDict({'train': train_ds, 'validation': val_ds, 'test': test_ds})
        
class DatasetDict(DsD):
    @staticmethod
    def load_from_disk(dataset_path: str) -> "DatasetDict":
        return load_from_disk(dataset_path=dataset_path)


from .doc import Doc, DocBin, Entity, Relation, Event
from .dataset import Dataset, DatasetDict
from .couplet import Couplet, CoupletBin



__all__ = ["Doc", "DocBin", "Entity", "Relation", "Event", "Dataset", "DatasetDict", "Couplet", "CoupletBin"]